"""Combine estimated DS phases with PS phases to form interferograms."""

from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from os import fspath
from pathlib import Path
from typing import Any, Iterable, Literal, Optional, Sequence, Union

import numpy as np
from numpy.typing import ArrayLike
from opera_utils import get_dates
from osgeo import gdal
from pydantic import BaseModel, Field, ValidationInfo, field_validator, model_validator
from scipy.ndimage import uniform_filter
from tqdm.contrib.concurrent import thread_map

from dolphin import io, utils
from dolphin._types import DateOrDatetime, Filename, T

gdal.UseExceptions()

logger = logging.getLogger(__name__)

DEFAULT_SUFFIX = ".int.vrt"


class VRTInterferogram(BaseModel, extra="allow"):
    """Create an interferogram using a VRTDerivedRasterBand.

    Attributes
    ----------
    ref_slc : Union[Path, str]
        Path to reference SLC file
    sec_slc : Union[Path, str]
        Path to secondary SLC file
    path : Optional[Path], optional
        Path to output interferogram. Defaults to Path('<date1>_<date2>.int.vrt'),
        placed in the same directory as `ref_slc`.
    outdir : Optional[Path], optional
        Directory to place output interferogram. Defaults to the same directory as
        `ref_slc`. If only `outdir` is specified, the output interferogram will
        be named '<date1>_<date2>.int.vrt', where the dates are parsed from the
        inputs. If `path` is specified, this is ignored.
    subdataset : Optional[str], optional
        Subdataset to use for the input files (if passing file paths
        to NETCDF/HDF5 files).
        Defaults to None.
    date_format : str, optional
        Date format to use when parsing dates from the input files.
        Defaults to '%Y%m%d'.
    pixel_function : str, optional
        GDAL Pixel function to use, choices={'cmul', 'mul'}.
        Defaults to 'cmul', which performs `ref_slc * sec_slc.conj()`.
        See https://gdal.org/drivers/raster/vrt.html#default-pixel-functions
    dates : tuple[datetime.date, datetime.date]
        Date of the interferogram (parsed from the input files).

    """

    subdataset: Optional[str] = Field(
        None,
        description="Subdataset to use for the input files. Defaults to None.",
    )
    ref_slc: Union[Path, str] = Field(..., description="Path to reference SLC file")
    sec_slc: Union[Path, str] = Field(..., description="Path to secondary SLC file")
    verify_slcs: bool = Field(
        True, description="Raise an error if `ref_slc` or `sec_slc` aren't readable."
    )
    outdir: Optional[Path] = Field(
        None,
        description=(
            "Directory to place output interferogram. Defaults to the same"
            " directory as `ref_slc`. If only `outdir` is specified, the output"
            f" interferogram will be named '<date1>_<date2>{DEFAULT_SUFFIX}', where the"
            " dates are parsed from the inputs. If `path` is specified, this is ignored"
        ),
        validate_default=True,
    )
    path: Optional[Path] = Field(
        None,
        description=(
            f"Path to output interferogram. Defaults to <date1>_<date2>{DEFAULT_SUFFIX}"
            ", where the dates are parsed from the input files, placed in the same "
            "directory as `ref_slc`."
        ),
        validate_default=True,
    )
    date_format: str = Field(
        io.DEFAULT_DATETIME_FORMAT,
        description="datetime format used to parse SLC filenames",
    )
    ref_date: Optional[DateOrDatetime] = Field(
        None,
        description=(
            "Reference date of the interferogram. If not specified,"
            "will be parsed from `ref_slc` using `date_format`."
        ),
    )
    sec_date: Optional[DateOrDatetime] = Field(
        None,
        description=(
            "Secondary date of the interferogram. If not specified,"
            "will be parsed from `sec_slc` using `date_format`."
        ),
    )
    resolve_paths: bool = Field(
        True, description="Resolve paths of `ref_slc`/`sec_slc` when saving the VRT"
    )
    use_relative: bool = Field(
        False, description='If true, use `relativeToVRT="1" in the VRT'
    )
    write: bool = Field(True, description="Write the VRT file to disk")

    pixel_function: Literal["cmul", "mul"] = "cmul"
    _template = """\
<VRTDataset rasterXSize="{xsize}" rasterYSize="{ysize}">
    <VRTRasterBand dataType="CFloat32" band="1" subClass="VRTDerivedRasterBand">
        <PixelFunctionType>{pixel_function}</PixelFunctionType>
        <SimpleSource>
            <SourceFilename relativeToVRT="{rel}">{ref_slc}</SourceFilename>
        </SimpleSource>
        <SimpleSource>
            <SourceFilename relativeToVRT="{rel}">{sec_slc}</SourceFilename>
        </SimpleSource>
    </VRTRasterBand>
</VRTDataset>
    """

    @field_validator("ref_slc", "sec_slc")
    @classmethod
    def _check_gdal_string(cls, v: Union[Path, str], info: ValidationInfo):
        subdataset = info.data.get("subdataset")
        # If we're using a subdataset, create a the GDAL-readable string
        return io.format_nc_filename(v, subdataset)

    @field_validator("outdir")
    @classmethod
    def _check_output_dir(cls, v, info: ValidationInfo):
        if v is not None:
            return Path(v)
        # If outdir is not set, use the directory of the reference SLC
        ref_slc = str(info.data.get("ref_slc"))
        return utils._get_path_from_gdal_str(ref_slc).parent

    @model_validator(mode="after")
    def _parse_dates(self) -> VRTInterferogram:
        # Get the dates from the input files if not provided
        if self.ref_date is None:
            d = get_dates(self.ref_slc, fmt=self.date_format)
            if not d:
                msg = f"No dates found in '{self.ref_slc}' like {self.date_format}"
                raise ValueError(msg)
            self.ref_date = d[0]
        if self.sec_date is None:
            d = get_dates(self.sec_slc, fmt=self.date_format)
            if not d:
                msg = f"No dates found in '{self.sec_slc}' like {self.date_format}"
                raise ValueError(msg)
            self.sec_date = d[0]

        return self

    @model_validator(mode="after")
    def _resolve_files(self) -> VRTInterferogram:
        """Check that the inputs are the same size and geotransform."""
        if not self.ref_slc or not self.sec_slc:
            # Skip validation if files are not set
            return self
        if self.resolve_paths:
            self.ref_slc = utils._resolve_gdal_path(self.ref_slc)  # type: ignore[assignment]
            self.sec_slc = utils._resolve_gdal_path(self.sec_slc)  # type: ignore[assignment]
        return self

    @model_validator(mode="after")
    def _validate_files(self) -> VRTInterferogram:
        # Only run this check if we care to validate the readability
        if not self.verify_slcs:
            return self

        ds1 = gdal.Open(fspath(self.ref_slc))
        ds2 = gdal.Open(fspath(self.sec_slc))
        xsize, ysize = ds1.RasterXSize, ds1.RasterYSize
        xsize2, ysize2 = ds2.RasterXSize, ds2.RasterYSize
        if xsize != xsize2 or ysize != ysize2:
            msg = f"Input files {self.ref_slc} and {self.sec_slc} are not the same size"
            raise ValueError(msg)
        gt1 = ds1.GetGeoTransform()
        gt2 = ds2.GetGeoTransform()
        if gt1 != gt2:
            msg = f"{self.ref_slc} and {self.sec_slc} have different GeoTransforms"
            raise ValueError(msg)

        return self

    @model_validator(mode="after")
    def _form_path(self) -> VRTInterferogram:
        """Create the filename (if not provided) from the provided SLCs."""
        if self.path is not None:
            return self

        if self.outdir is None:
            # If outdir is not set, use the directory of the reference SLC
            self.outdir = utils._get_path_from_gdal_str(self.ref_slc).parent
        assert self.ref_date is not None
        assert self.sec_date is not None
        date_str = utils.format_date_pair(
            self.ref_date, self.sec_date, fmt=self.date_format
        )
        path = self.outdir / (date_str + DEFAULT_SUFFIX)
        self.path = path
        return self

    @model_validator(mode="after")
    def _write_vrt(self) -> VRTInterferogram:
        """Write out the VRT if requested."""
        if not self.write:
            return self
        assert self.path is not None
        if Path(self.path).exists():
            logger.info(f"Removing {self.path}")
            self.path.unlink()

        xsize, ysize = io.get_raster_xysize(self.ref_slc)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.path, "w") as f:
            f.write(
                self._template.format(
                    xsize=xsize,
                    ysize=ysize,
                    ref_slc=self.ref_slc,
                    sec_slc=self.sec_slc,
                    pixel_function=self.pixel_function,
                    rel=self.use_relative,
                )
            )
        io.copy_projection(self.ref_slc, self.path)
        return self

    def load(self):
        """Load the interferogram as a numpy array."""
        return io.load_gdal(self.path)

    @property
    def shape(self):  # noqa: D102
        xsize, ysize = io.get_raster_xysize(self.path)
        return (ysize, xsize)

    @property
    def dates(self):  # noqa: D102
        return (self.ref_date, self.sec_date)

    @classmethod
    def from_vrt_file(cls, path: Filename) -> VRTInterferogram:
        """Load a VRTInterferogram from an existing VRT file.

        Parameters
        ----------
        path : Filename
            Path to VRT file.

        Returns
        -------
        VRTInterferogram
            VRTInterferogram object.

        """
        from dolphin.io._readers import _parse_vrt_file

        # Use the parsing function
        (ref_slc, sec_slc), subdataset = _parse_vrt_file(path)
        if subdataset is not None:
            ref_slc = io.format_nc_filename(ref_slc, subdataset)
            sec_slc = io.format_nc_filename(sec_slc, subdataset)

        return cls.model_construct(
            ref_slc=ref_slc,
            sec_slc=sec_slc,
            path=Path(path).resolve(),
            subdataset=subdataset,
            date_format="%Y%m%d",
        )


# Alias for the pairs of filenames in ifgs
IfgPairT = tuple[Filename, Filename]


@dataclass
class Network:
    """A network of interferograms from a list of SLCs.

    Attributes
    ----------
    slc_list : list[Filename]
        list of SLCs to use to form interferograms.
    ifg_list : list[tuple[Filename, Filename]]
        list of `VRTInterferogram`s created from the SLCs.
    max_bandwidth : int | None, optional
        Maximum number of SLCs to include in an interferogram, by index distance.
        Defaults to None.
    max_temporal_baseline : Optional[float], default = None
        Maximum temporal baseline to include in an interferogram, in days.
    include_annual : bool, default = False
        Attempt to include annual pairs, in addition to the other type of network
        requested. Will skip if no pairs exist within 365 +/- `buffer_days`
    annual_buffer_days : float, default = 30
        Search buffer for finding annual pairs, if `include_annual=True`
    date_format : str, optional
        Date format to use when parsing dates from the input files (only
        used if setting `max_temporal_baseline`).
        defaults to '%Y%m%d'.
    dates: Sequence[DateOrDatetime], optional
        Alternative to `date_format`: manually specify the date/datetime of each item in
        `slc_list` instead of parsing the name.
        Only used for `max_temporal_baseline` networks.
    reference_idx : int | None, optional
        Index of the SLC to use as the reference for all interferograms.
        Defaults to None.
    indexes : Sequence[tuple[int, int]], optional
        Manually list (ref_idx, sec_idx) pairs to use for interferograms.
    subdataset : Optional[str], default = None
        If passing NetCDF files in `slc_list, the subdataset of the image data
        within the file.
        Can also pass a sequence of one subdataset per entry in `slc_list`
    verify_slcs : bool, default = True
        Raise an error if any SLCs aren't GDAL-readable.
    write : bool
        Whether to write the VRT files to disk. Defaults to True.

    """

    slc_list: Sequence[Filename]
    outdir: Optional[Filename] = None
    max_bandwidth: Optional[int] = None
    max_temporal_baseline: Optional[float] = None
    include_annual: bool = False
    annual_buffer_days: float = 30
    date_format: str = io.DEFAULT_DATETIME_FORMAT
    dates: Optional[Sequence[DateOrDatetime]] = None
    reference_idx: Optional[int] = None
    indexes: Optional[Sequence[tuple[int, int]]] = None
    subdataset: Optional[Union[str, Sequence[str]]] = None
    verify_slcs: bool = True
    resolve_paths: bool = True
    use_relative: bool = False
    write: bool = True

    def __post_init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        if self.subdataset is None or isinstance(self.subdataset, str):
            self._slc_to_subdataset = {slc: self.subdataset for slc in self.slc_list}
        else:
            # We're passing a sequence
            assert len(self.subdataset) == len(self.slc_list)
            self._slc_to_subdataset = dict(zip(self.slc_list, self.subdataset))

        if self.outdir is None:
            self.outdir = Path(self.slc_list[0]).parent

        # Set the dates to be used for each ifg
        if self.dates is None:
            # Use the first one we find in the name
            self.dates = [get_dates(f, fmt=self.date_format)[0] for f in self.slc_list]
        if len(self.dates) != len(self.slc_list):
            msg = f"{len(self.dates) = }, but {len(self.slc_list) = }"
            raise ValueError(msg)
        self._slc_to_date = dict(zip(self.slc_list, self.dates))

        # Run the appropriate network creation based on the options we passed
        self.slc_file_pairs = self._make_ifg_pairs()

        # Create each VRT file
        self.ifg_list: list[VRTInterferogram] = self._create_vrt_ifgs()

    def _make_ifg_pairs(self) -> list[IfgPairT]:
        """Form interferogram pairs from a list of SLC files sorted by date."""
        assert self.dates is not None
        ifgs: list[IfgPairT] = []
        # For each type of possible network, add the ifgs to the list
        if self.indexes is not None:
            # Give the option to select exactly which interferograms to create
            ifgs.extend(
                [
                    (self.slc_list[ref_idx], self.slc_list[sec_idx])
                    for ref_idx, sec_idx in self.indexes
                ]
            )
        if self.max_bandwidth is not None:
            ifgs.extend(Network.limit_by_bandwidth(self.slc_list, self.max_bandwidth))
        if self.max_temporal_baseline is not None:
            ifgs.extend(
                Network.limit_by_temporal_baseline(
                    self.slc_list,
                    dates=self.dates,
                    max_temporal_baseline=self.max_temporal_baseline,
                )
            )
        if self.reference_idx is not None:
            ifgs.extend(
                Network._single_reference_network(self.slc_list, self.reference_idx)
            )

        if self.include_annual:
            # Add in the annual pairs, then re-sort
            annual_ifgs = Network.find_annuals(
                self.slc_list, self.dates, buffer_days=self.annual_buffer_days
            )
            ifgs.extend(annual_ifgs)

        if not ifgs:
            msg = "No valid ifg list generation method specified"
            raise ValueError(msg)

        # Sort and dedupe them
        return sorted(set(ifgs))

    def _create_vrt_ifgs(self) -> list[VRTInterferogram]:
        """Write out a VRTInterferogram for each ifg."""
        ifg_list: list[VRTInterferogram] = []
        date_pairs = self._get_ifg_date_pairs()
        for idx, (ref, sec) in enumerate(self._gdal_file_strings):
            ref_date, sec_date = date_pairs[idx]

            v = VRTInterferogram(
                ref_slc=ref,
                sec_slc=sec,
                date_format=self.date_format,
                outdir=self.outdir,
                ref_date=ref_date,
                sec_date=sec_date,
                verify_slcs=self.verify_slcs,
                resolve_paths=self.resolve_paths,
                use_relative=self.use_relative,
                write=self.write,
            )
            ifg_list.append(v)
        return ifg_list

    @property
    def _gdal_file_strings(self):
        # format each file in each pair
        out = []
        for slc1, slc2 in self.slc_file_pairs:
            sd1 = self._slc_to_subdataset[slc1]
            sd2 = self._slc_to_subdataset[slc2]
            out.append(
                [io.format_nc_filename(slc1, sd1), io.format_nc_filename(slc2, sd2)]
            )
        return out

    def _get_ifg_date_pairs(self) -> list[tuple[DateOrDatetime, DateOrDatetime]]:
        date_pairs = []
        for slc1, slc2 in self.slc_file_pairs:
            d1 = self._slc_to_date[slc1]
            d2 = self._slc_to_date[slc2]
            date_pairs.append((d1, d2))
        return date_pairs

    def __repr__(self):
        return (
            f"Network(ifg_list={self.ifg_list}, slc_list={self.slc_list},"
            f" max_bandwidth={self.max_bandwidth},"
            f" max_temporal_baseline={self.max_temporal_baseline},"
            f" reference_idx={self.reference_idx})"
        )

    def __str__(self):
        return (
            f"Network of {len(self.ifg_list)} interferograms, "
            f"max_bandwidth={self.max_bandwidth}, "
            f"max_temporal_baseline={self.max_temporal_baseline}, "
            f"reference_idx={self.reference_idx}"
        )

    @staticmethod
    def _single_reference_network(
        slc_list: Sequence[Filename], reference_idx=0
    ) -> list[IfgPairT]:
        """Form a list of single-reference interferograms."""
        if len(slc_list) < 2:
            msg = "Need at least two dates to make an interferogram list"
            raise ValueError(msg)
        ref = slc_list[reference_idx]
        return [tuple(sorted([ref, date])) for date in slc_list if date != ref]

    @staticmethod
    def limit_by_bandwidth(
        slc_list: Iterable[T], max_bandwidth: int
    ) -> list[tuple[T, T]]:
        """Form a list of the "nearest-`max_bandwidth`" ifgs.

        Parameters
        ----------
        slc_list : Iterable[Filename]
            list of dates of SLCs
        max_bandwidth : int
            Largest allowed span of ifgs, by index distance, to include.
            max_bandwidth=1 will only include nearest-neighbor ifgs.

        Returns
        -------
        list
            Pairs of (date1, date2) ifgs

        """
        slc_to_idx = {s: idx for idx, s in enumerate(slc_list)}
        return [
            (a, b)
            for (a, b) in Network.all_pairs(slc_list)
            if slc_to_idx[b] - slc_to_idx[a] <= max_bandwidth
        ]

    @staticmethod
    def limit_by_temporal_baseline(
        slc_list: Iterable[T],
        dates: Sequence[DateOrDatetime],
        max_temporal_baseline: Optional[float] = None,
    ) -> list[tuple[T, T]]:
        """Form a list of the ifgs limited to a maximum temporal baseline.

        Parameters
        ----------
        slc_list : Iterable[Filename]
            Iterable of input SLC files
        dates: Sequence[DateOrDatetime]
            Dates or Datetimes corresponding to `slc_list`
        max_temporal_baseline : float, optional
            Largest allowed span of ifgs, by index distance, to include.
            max_bandwidth=1 will only include nearest-neighbor ifgs.

        Returns
        -------
        list
            Pairs of (date1, date2) ifgs

        Raises
        ------
        ValueError
            If any of the input files have more than one date.

        """
        ifg_strs = Network.all_pairs(slc_list)
        ifg_dates = Network.all_pairs(dates)
        baselines = [Network.temporal_baseline(ifg) for ifg in ifg_dates]
        return [
            ifg for ifg, b in zip(ifg_strs, baselines) if b <= max_temporal_baseline
        ]

    @staticmethod
    def find_annuals(
        slc_list: Iterable[T],
        dates: Sequence[DateOrDatetime],
        buffer_days: float = 30,
    ) -> list[tuple[T, T]]:
        """Pick out interferograms which are closest to 1 year in span.

        We only want to pick 1 ifg per date, closest to a year, but we will skip
        a date if it doesn't have an ifg of baseline 365 +/- `buffer_days`.
        """
        # keep track how far into ifg_list the last sar date was
        date_to_date_pair: dict[
            DateOrDatetime, tuple[DateOrDatetime, DateOrDatetime]
        ] = {}
        date_to_file: dict[DateOrDatetime, tuple[T, T]] = {}
        slc_pairs = Network.all_pairs(slc_list)
        date_pairs = Network.all_pairs(dates)
        for ifg, date_pair in zip(slc_pairs, date_pairs):
            early = date_pair[0]
            baseline_days = Network.temporal_baseline(date_pair)
            if abs(baseline_days - 365) > buffer_days:
                continue
            dp = date_to_date_pair.get(early)
            # Use this ifg as the annual if none exist, or if it's closer to 365
            if dp is None or abs(baseline_days - 365) < Network.temporal_baseline(dp):
                date_to_file[early] = ifg
        return sorted(date_to_file.values())

    @staticmethod
    def all_pairs(slc_list: Iterable[T]) -> list[tuple[T, T]]:
        """Create the list of all possible ifg pairs from slc_list."""
        return list(itertools.combinations(slc_list, r=2))

    @staticmethod
    def temporal_baseline(ifg_pair: tuple[DateOrDatetime, DateOrDatetime]):
        """Compute the temporal baseline (in decimal days) between the date pair."""
        return (ifg_pair[1] - ifg_pair[0]).total_seconds() / 86400

    def __len__(self):
        return len(self.ifg_list)

    def __getitem__(self, idx):
        return self.ifg_list[idx]

    def __iter__(self):
        return iter(self.ifg_list)

    def __contains__(self, item):
        return item in self.ifg_list

    def __eq__(self, other):
        return self.ifg_list == other.ifg_list


def estimate_correlation_from_phase(
    ifg: Union[VRTInterferogram, ArrayLike], window_size: Union[int, tuple[int, int]]
) -> np.ndarray:
    """Estimate correlation from only an interferogram (no SLCs/magnitudes).

    This is a simple correlation estimator that takes the (complex) average
    in a moving window in an interferogram. Used to get some estimate of interferometric
    correlation on the result of phase-linking interferograms.

    Parameters
    ----------
    ifg : Union[VRTInterferogram, ArrayLike]
        Interferogram to estimate correlation from.
        If a VRTInterferogram, will load and take the phase.
        If `ifg` is complex, will normalize to unit magnitude before estimating.
    window_size : Union[int, tuple[int, int]]
        Size of window to use for correlation estimation.
        If int, will use a square window of that size.
        If tuple, the rectangular window has shape  `size=(row_size, col_size)`.

    Returns
    -------
    np.ndarray
        Correlation array

    """
    if isinstance(ifg, VRTInterferogram):
        ifg = ifg.load()
    nan_mask = np.isnan(ifg)
    zero_mask = ifg == 0
    if not np.iscomplexobj(ifg):
        # If they passed phase, convert to complex
        inp = np.exp(1j * np.nan_to_num(ifg))
    else:
        # If they passed complex, normalize to unit magnitude
        inp = np.exp(1j * np.nan_to_num(np.angle(ifg)))

    # Note: the clipping is from possible partial windows producing correlation
    # above 1
    cor = np.clip(np.abs(uniform_filter(inp, window_size, mode="nearest")), 0, 1)
    # Return the input nans to nan
    cor[nan_mask] = np.nan
    # If the input was 0, the correlation is 0
    cor[zero_mask] = 0
    return cor


def estimate_interferometric_correlations(
    ifg_filenames: Sequence[Filename],
    window_size: tuple[int, int],
    out_driver: str = "GTiff",
    out_suffix: str = ".cor.tif",
    options: Sequence[str] = io.DEFAULT_TIFF_OPTIONS,
    keep_bits: int = 10,
    num_workers: int = 3,
) -> list[Path]:
    """Estimate correlations for a sequence of interferograms.

    Will use the same filename base as inputs with a new suffix.

    Parameters
    ----------
    ifg_filenames : Sequence[Filename]
        Paths to complex interferogram files.
    window_size : tuple[int, int]
        (row, column) size of window to use for estimate.
    out_driver : str, optional
        Name of output GDAL driver, by default "GTiff".
    out_suffix : str, optional
        File suffix to use for correlation files, by default ".cor.tif"
    options : list[str], optional
        GDAL Creation options for the output array.
    keep_bits : int, optional
        Number of bits to preserve in mantissa. Defaults to None.
        Lower numbers will truncate the mantissa more and enable more compression.
    num_workers : int
        Number of threads to use for stitching in parallel.
        Default = 3

    Returns
    -------
    list[Path]
        Paths to newly written correlation files.

    """
    path_tuples: list[tuple[Path, Path]] = []
    output_paths: list[Path] = []
    for fn in ifg_filenames:
        ifg_path = Path(fn)
        cor_path = ifg_path.with_suffix(out_suffix)
        output_paths.append(cor_path)
        if cor_path.exists():
            logger.info(f"Skipping existing interferometric correlation for {ifg_path}")
            continue
        path_tuples.append((ifg_path, cor_path))

    def process_ifg(args):
        ifg_path, cor_path = args
        logger.debug(f"Estimating correlation for {ifg_path}, writing to {cor_path}")
        ifg = io.load_gdal(ifg_path)
        cor = estimate_correlation_from_phase(ifg, window_size=window_size)
        if keep_bits:
            io.round_mantissa(cor, keep_bits=keep_bits)
        io.write_arr(
            arr=cor,
            output_name=cor_path,
            like_filename=ifg_path,
            driver=out_driver,
            options=options,
        )

    thread_map(
        process_ifg,
        path_tuples,
        max_workers=num_workers,
        desc="Estimating correlations",
    )

    return output_paths


def _create_vrt_conj(
    filename: Filename, output_filename: Filename, is_relative: bool = False
):
    """Create a VRT raster of the conjugate of `filename`."""
    xsize, ysize = io.get_raster_xysize(filename)
    rel = "1" if is_relative else "0"

    # See https://gdal.org/drivers/raster/vrt.html#default-pixel-functions
    vrt_template = f"""\
<VRTDataset rasterXSize="{xsize}" rasterYSize="{ysize}">
    <VRTRasterBand dataType="CFloat32" band="1" subClass="VRTDerivedRasterBand">
        <PixelFunctionType>conj</PixelFunctionType>
        <SimpleSource>
            <SourceFilename relativeToVRT="{rel}">{filename}</SourceFilename>
        </SimpleSource>
    </VRTRasterBand>
</VRTDataset>
    """
    with open(output_filename, "w") as f:
        f.write(vrt_template)
    io.copy_projection(filename, output_filename)


def convert_pl_to_ifg(
    phase_linked_slc: Filename,
    reference_date: DateOrDatetime,
    output_dir: Filename,
    dry_run: bool = False,
) -> Path:
    """Convert a phase-linked SLC to an interferogram by conjugating the phase.

    The SLC has already been multiplied by the (conjugate) phase of a reference SLC,
    so it only needs to be conjugated to put it in the form (ref * sec.conj()).

    Parameters
    ----------
    phase_linked_slc : Filename
        Path to phase-linked SLC file.
    reference_date : DateOrDatetime
        Reference date of the interferogram.
    output_dir : Filename
        Directory to place the renamed file.
    dry_run : bool
        Flag indicating that the new ifgs shouldn't be written to disk.
        Default = False (the ifgs will be created/written to disk.)
        `dry_run=True` is used to plan out which ifgs will be formed
        before actually running the workflow.

    Returns
    -------
    Path
        Path to renamed file.

    """
    # The phase_linked_slc will be named with the secondary date.
    # Make the output from that, plus the given reference date
    secondary_date = get_dates(phase_linked_slc)[-1]
    date_str = utils.format_date_pair(reference_date, secondary_date)
    out_name = Path(output_dir) / f"{date_str}.int.vrt"
    if dry_run:
        return out_name
    out_name.parent.mkdir(parents=True, exist_ok=True)
    # Now make a VRT to do the .conj
    _create_vrt_conj(phase_linked_slc, out_name)
    return out_name



import warnings
from pathlib import Path

import numpy as np
import pytest
import rasterio as rio
from rasterio.errors import NotGeoreferencedWarning

from dolphin.io._core import load_gdal
from dolphin.io._writers import (
    BackgroundBlockWriter,
    BackgroundRasterWriter,
    BackgroundStackWriter,
    RasterWriter,
)

# Note: uses the fixtures from conftest.py

# Get combinations of slices
slices_to_test = [slice(None), 1, slice(0, 10, 2)]


# Filter rasterio georeferencing warnings
@pytest.fixture(autouse=True)
def suppress_not_georeferenced_warning():
    """
    Pytest fixture to suppress NotGeoreferencedWarning in tests.

    This fixture automatically applies to all test functions in the module
    where it's defined, suppressing the specified warning.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=NotGeoreferencedWarning)
        yield


@pytest.fixture
def output_file_list(slc_file_list):
    suffix = Path(slc_file_list[0]).suffix
    return [f"{f}_out{suffix}" for f in slc_file_list]


def test_background_block_writer(output_file_list, slc_file_list):
    from dolphin.io import write_arr

    write_arr(arr=None, output_name=output_file_list[0], like_filename=slc_file_list[0])
    data = np.random.randn(5, 10)
    w = BackgroundBlockWriter()
    rows, cols = slice(0, 5), slice(0, 10)
    w.queue_write(data, output_file_list[0], 0, 0)
    # Make sure we dont write too late
    w.notify_finished()
    assert w._thread.is_alive() is False

    assert np.allclose(load_gdal(output_file_list[0], rows=rows, cols=cols), data)


# #### RasterReader Tests ####
class TestRasterWriter:
    def raster_init(self, slc_file_list, slc_stack):
        # ignore georeferencing warnings
        with pytest.warns(rio.errors.NotGeoreferencedWarning):
            w = RasterWriter(slc_file_list[0])
        assert w.shape == slc_stack[0].shape
        assert w.dtype == slc_stack[0].dtype
        assert w.ndim == 2
        assert w.dtype == np.complex64
        assert w.filename == slc_file_list[0]
        assert w.band == 1
        assert w.closed is False
        w.close()

    def test_write(self, slc_file_list):
        data = np.random.randn(5, 10)
        w = RasterWriter(slc_file_list[0])
        rows, cols = slice(0, 5), slice(0, 10)
        w[rows, cols] = data
        w.close()

        assert np.allclose(load_gdal(slc_file_list[0], rows=rows, cols=cols), data)

    def test_context_manager(self, slc_file_list):
        rows, cols = slice(0, 5), slice(0, 10)
        data = np.random.randn(5, 10)
        with RasterWriter(slc_file_list[0]) as w:
            w[rows, cols] = data
            assert w.closed is False

        assert w.closed is True
        assert np.allclose(load_gdal(slc_file_list[0], rows=rows, cols=cols), data)


class TestBackgroundRasterWriter:
    def test_init(self, slc_file_list):
        brw = BackgroundRasterWriter(slc_file_list[0])
        with rio.open(slc_file_list[0]) as src:
            assert brw.shape == src.shape
            assert brw.dtype == src.dtypes[0]
        assert brw.ndim == 2
        assert brw.closed is False
        brw.close()
        assert brw.closed is True

    def test_setitem(self, slc_file_list):
        data = np.random.randn(5, 10)
        w = BackgroundRasterWriter(slc_file_list[0])
        rows, cols = slice(0, 5), slice(0, 10)
        w[rows, cols] = data
        # Make sure we dont write too late
        w.close()
        assert w._thread.is_alive() is False

        assert np.allclose(load_gdal(slc_file_list[0], rows=rows, cols=cols), data)

    def test_context_manager(self, slc_file_list):
        rows, cols = slice(0, 5), slice(0, 10)
        data = np.random.randn(5, 10)
        with BackgroundRasterWriter(slc_file_list[0]) as w:
            w[rows, cols] = data
            assert w.closed is False

        assert w.closed is True
        assert w._thread.is_alive() is False
        assert np.allclose(load_gdal(slc_file_list[0], rows=rows, cols=cols), data)


class TestBackgroundStackWriter:
    def test_stack(self, slc_file_list, output_file_list):
        w = BackgroundStackWriter(output_file_list, like_filename=slc_file_list[0])
        with rio.open(output_file_list[0]) as src:
            shape_3d = (len(slc_file_list), *src.shape)
            assert w.shape == shape_3d
            assert w.dtype == src.dtypes[0]

        assert w.ndim == 3
        assert w.closed is False
        w.close()
        assert w.closed is True

    def test_setitem(self, output_file_list, slc_file_list):
        data = np.random.randn(len(output_file_list), 5, 10)
        w = BackgroundStackWriter(output_file_list, like_filename=slc_file_list[0])
        rows, cols = slice(0, 5), slice(0, 10)
        w[:, rows, cols] = data
        # Make sure we dont write too late
        w.close()
        assert w._thread.is_alive() is False

        for i in range(len(output_file_list)):
            assert np.allclose(
                load_gdal(output_file_list[i], rows=rows, cols=cols), data[i]
            )



from __future__ import annotations

import datetime
import itertools
import logging
import multiprocessing as mp
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Optional, Union

import numpy as np
import xarray
from pyproj import CRS, Transformer
from scipy.interpolate import RegularGridInterpolator as Interpolator

from dolphin._types import Filename, TropoModel

logger = logging.getLogger(__name__)

###########
# Mostly inherited from RAiDER

__all__ = ["delay_from_netcdf"]


def delay_from_netcdf(
    dt: datetime.datetime,
    weather_model_file: Filename,
    SNWE: tuple[float, float, float, float],
    height_levels: Optional[list[float]],
    out_proj: Union[int, str] = 4326,
    weather_model: TropoModel = TropoModel.ECMWF,
):
    """Calculate integrated delays on query points.

    Parameters
    ----------
    dt: Datetime
        Datetime object for determining when to calculate delays
    weather_model_file: string
        Name of the NETCDF file containing a pre-processed weather model
    SNWE: tuple[float]
        Bounding box in lat/lon
    height_levels: list
        (optional) list of height levels on which to calculate delays.
        Only needed for cube generation.
    out_proj: int,str
        (optional) EPSG code for output projection
    weather_model: TropoModel
        weather model  (ECMWF or HRES)

    Returns
    -------
    xarray Dataset *or* ndarrays:
        wet and hydrostatic delays at the grid nodes / query points.

    """
    # Load CRS from weather model file
    with xarray.load_dataset(weather_model_file) as ds:
        # The dataset in the original netcdf that is download includes:
        # t, z, q and lnsp
        # The converted netcdf to the format usable for our purpose consists of
        # calculated datasets of hydro and wet. so in the following we check for
        # the netcdf file to include those data and if not, we convert it
        if "hydro" in ds:
            wmodel_file = weather_model_file
        else:
            wmodel_file = convert_netcdf(
                weather_model_file=weather_model_file,
                lat_lon_bounds=SNWE,
                weather_model_output_dir="./tmp",
                weather_model=weather_model,
            )

        try:
            wm_proj = CRS.from_wkt(ds["proj"].attrs["crs_wkt"])
        except KeyError:
            logger.debug(
                "I can't find a CRS in the weather model file, "
                "so I will assume you are using WGS84"
            )
            wm_proj = CRS.from_epsg(4326)

    # get heights
    with xarray.load_dataset(wmodel_file) as ds:
        wm_levels = ds.z.values

    if height_levels is None:
        height_levels = wm_levels

    ds = _get_delays_on_cube(
        dt, wmodel_file, wm_proj, SNWE, height_levels, out_epsg=out_proj
    )

    return ds


def transform_bbox(snwe_in, dest_crs=4326, src_crs=4326, margin=100.0):
    """Transform bbox to lat/lon or another CRS for use with rest of workflow.

    Returns: SNWE.
    """
    # TODO - Handle dateline crossing
    if isinstance(src_crs, int):
        src_crs = CRS.from_epsg(src_crs)
    elif isinstance(src_crs, str):
        src_crs = CRS(src_crs)

    # Handle margin for input bbox in degrees
    if src_crs.axis_info[0].unit_name == "degree":
        margin = margin / 1.0e5

    if isinstance(dest_crs, int):
        dest_crs = CRS.from_epsg(dest_crs)
    elif isinstance(dest_crs, str):
        dest_crs = CRS(dest_crs)

    # If dest_crs is same as src_crs
    if dest_crs == src_crs:
        return snwe_in

    T = Transformer.from_crs(src_crs, dest_crs, always_xy=True)
    xs = np.linspace(snwe_in[2] - margin, snwe_in[3] + margin, num=11)
    ys = np.linspace(snwe_in[0] - margin, snwe_in[1] + margin, num=11)
    X, Y = np.meshgrid(xs, ys)

    # Transform to lat/lon
    xx, yy = T.transform(X, Y)

    # query_area convention
    snwe = [np.nanmin(yy), np.nanmax(yy), np.nanmin(xx), np.nanmax(xx)]
    return snwe


def get_output_xygrid(
    out_proj: CRS,
    SNWE: tuple[float, float, float, float],
    cube_spacing_m: float = 500,
) -> tuple[np.ndarray, np.ndarray]:
    """Read the SNWE bounding box in lat/lon and create the output grid.

    Parameters
    ----------
    SNWE: tuple[float, float, float, float]
        The bounding box in lat/lon coordinates
    cube_spacing_m: float
        The output grid spacing in meters
    out_proj: CRS
        The output grid projection

    Returns
    -------
        The x and y coordinates of the output grid

    """
    out_snwe = transform_bbox(SNWE, src_crs=4326, dest_crs=out_proj)
    logger.debug(f"Output SNWE: {out_snwe}")

    # Build the output grid
    if all(axis_info.unit_name == "degree" for axis_info in out_proj.axis_info):
        out_spacing = cube_spacing_m / 1e5
    else:
        out_spacing = cube_spacing_m

    xpts = np.arange(out_snwe[2], out_snwe[3] + out_spacing, out_spacing)
    ypts = np.arange(out_snwe[1], out_snwe[0] - out_spacing, -out_spacing)
    return xpts, ypts


def _get_delays_on_cube(dt, weather_model_file, wm_proj, SNWE, heights, out_epsg):
    """Generate delays on cube."""
    zpts = np.array(heights)
    out_type = "zenith"

    # Get ZTD interpolators
    try:
        ifWet, ifHydro = _get_interpolators(weather_model_file, "total")
    except RuntimeError:
        logger.exception(
            "Failed to get weather model %s interpolators.", weather_model_file
        )

    # Build cube
    crs = CRS.from_epsg(out_epsg)
    xpts, ypts = get_output_xygrid(SNWE=SNWE, out_proj=crs)
    wetDelay, hydroDelay = _build_cube(xpts, ypts, zpts, wm_proj, crs, [ifWet, ifHydro])

    if np.isnan(wetDelay).any() or np.isnan(hydroDelay).any():
        logger.critical("There are missing delay values. Check your inputs.")

    # Write output file
    ds = _write_results_to_xarray(
        dt, xpts, ypts, zpts, crs, wetDelay, hydroDelay, weather_model_file, out_type
    )

    return ds


def _get_interpolators(wm_file, kind="pointwise", shared=False):
    """Read 3D gridded data from a processed weather model file.

    and wrap it with the scipy RegularGridInterpolator.
    The interpolator grid is (y, x, z)
    """
    # Get the weather model data
    try:
        ds = xarray.load_dataset(wm_file)
    except ValueError:
        ds = wm_file

    xs_wm = np.array(ds.variables["x"][:])
    ys_wm = np.array(ds.variables["y"][:])
    zs_wm = np.array(ds.variables["z"][:])

    wet = ds.variables["wet_total" if kind == "total" else "wet"][:]
    hydro = ds.variables["hydro_total" if kind == "total" else "hydro"][:]

    wet = np.array(wet).transpose(1, 2, 0)
    hydro = np.array(hydro).transpose(1, 2, 0)

    if np.any(np.isnan(wet)) or np.any(np.isnan(hydro)):
        logger.critical("Weather model contains NaNs!")

    # If shared interpolators are requested
    # The arrays are not modified - so turning off lock for performance
    if shared:
        xs_wm = _make_shared_raw(xs_wm)
        ys_wm = _make_shared_raw(ys_wm)
        zs_wm = _make_shared_raw(zs_wm)
        wet = _make_shared_raw(wet)
        hydro = _make_shared_raw(hydro)

    ifWet = Interpolator(
        (ys_wm, xs_wm, zs_wm), wet, fill_value=np.nan, bounds_error=False
    )
    ifHydro = Interpolator(
        (ys_wm, xs_wm, zs_wm), hydro, fill_value=np.nan, bounds_error=False
    )

    return ifWet, ifHydro


def _make_shared_raw(inarr):
    """Make numpy view array of mp.Array."""
    # Create flat shared array

    shared_arr = mp.RawArray("d", inarr.size)
    # Create a numpy view of it
    shared_arr_np = np.ndarray(inarr.shape, dtype=np.float64, buffer=shared_arr)
    # Copy data to shared array
    np.copyto(shared_arr_np, inarr)

    return shared_arr_np


def _build_cube(xpts, ypts, zpts, model_crs, pts_crs, interpolators):
    """Iterate over interpolators and build a cube using Zenith."""
    # Create a regular 2D grid
    xx, yy = np.meshgrid(xpts, ypts)

    # Output arrays
    outputArrs = [
        np.zeros((zpts.size, ypts.size, xpts.size)) for mm in range(len(interpolators))
    ]

    # Loop over heights and compute delays
    for ii, ht in enumerate(zpts):
        # pts is in weather model system;
        if model_crs != pts_crs:
            # lat / lon / height for hrrr
            pts = transform_points(yy, xx, np.full(yy.shape, ht), pts_crs, model_crs)
        else:
            pts = np.stack([yy, xx, np.full(yy.shape, ht)], axis=-1)

        for mm, intp in enumerate(interpolators):
            outputArrs[mm][ii, ...] = intp(pts)

    return outputArrs


def transform_points(
    lats: np.ndarray, lons: np.ndarray, hgts: np.ndarray, old_proj: CRS, new_proj: CRS
) -> np.ndarray:
    """Transform lat/lon/hgt data to an array of points in a new projection.

    Parameters
    ----------
    lats : ndarray
        WGS-84 latitude (EPSG: 4326)
    lons : ndarray
        ditto for longitude
    hgts : ndarray
        Ellipsoidal height in meters
    old_proj: CRS
        the original projection of the points
    new_proj: CRS
        the new projection in which to return the points

    Returns
    -------
    ndarray: the array of query points in the weather model coordinate system (YX)

    """
    # Flags for flipping inputs or outputs
    if not isinstance(new_proj, CRS):
        new_proj = CRS.from_epsg(new_proj.lstrip("EPSG:"))
    if not isinstance(old_proj, CRS):
        old_proj = CRS.from_epsg(old_proj.lstrip("EPSG:"))

    t = Transformer.from_crs(old_proj, new_proj, always_xy=True)

    res = t.transform(lons, lats, hgts)

    # lat/lon/height
    return np.stack([res[1], res[0], res[2]], axis=-1)


def _write_results_to_xarray(
    dt: datetime.datetime,
    xpts: np.ndarray,
    ypts: np.ndarray,
    zpts: np.ndarray,
    crs: CRS,
    wetDelay: np.ndarray,
    hydroDelay: np.ndarray,
    weather_model_file: str,
    out_type: str,
):
    """Write a 1-D array to a NETCDF5 file."""
    ds = xarray.Dataset(
        data_vars={
            "wet": (
                ["z", "y", "x"],
                wetDelay,
                {
                    "units": "m",
                    "description": f"wet {out_type} delay",
                    # 'crs': crs.to_epsg(),
                    "grid_mapping": "crs",
                },
            ),
            "hydro": (
                ["z", "y", "x"],
                hydroDelay,
                {
                    "units": "m",
                    # 'crs': crs.to_epsg(),
                    "description": f"hydrostatic {out_type} delay",
                    "grid_mapping": "crs",
                },
            ),
        },
        coords={
            "x": (["x"], xpts),
            "y": (["y"], ypts),
            "z": (["z"], zpts),
        },
        attrs={
            "Conventions": "CF-1.7",
            "title": "Geo cube",
            "source": Path(weather_model_file).name,
            "history": str(datetime.datetime.now(datetime.timezone.utc)),
            "description": f"Geo cube - {out_type}",
            "reference_time": dt.strftime("%Y%m%dT%H:%M:%S"),
        },
    )

    # Write projection system mapping
    ds["crs"] = int(-2147483647)  # dummy placeholder
    for k, v in crs.to_cf().items():
        ds.crs.attrs[k] = v

    # Write z-axis information
    ds.z.attrs["axis"] = "Z"
    ds.z.attrs["units"] = "m"
    ds.z.attrs["description"] = "height above ellipsoid"

    # If in degrees
    if crs.axis_info[0].unit_name == "degree":
        ds.y.attrs["units"] = "degrees_north"
        ds.y.attrs["standard_name"] = "latitude"
        ds.y.attrs["long_name"] = "latitude"

        ds.x.attrs["units"] = "degrees_east"
        ds.x.attrs["standard_name"] = "longitude"
        ds.x.attrs["long_name"] = "longitude"

    else:
        ds.y.attrs["axis"] = "Y"
        ds.y.attrs["standard_name"] = "projection_y_coordinate"
        ds.y.attrs["long_name"] = "y-coordinate in projected coordinate system"
        ds.y.attrs["units"] = "m"

        ds.x.attrs["axis"] = "X"
        ds.x.attrs["standard_name"] = "projection_x_coordinate"
        ds.x.attrs["long_name"] = "x-coordinate in projected coordinate system"
        ds.x.attrs["units"] = "m"

    return ds


def group_netcdf_by_date(
    files: Iterable[Filename],
    file_date_fmt: str = "datetime64[s]",
) -> dict[datetime.datetime, list[Filename]]:
    """Combine files by date and time into a dict.

    Parameters
    ----------
    files : Iterable[Filename]
        Path to folder containing files with dates in the time dataset.
    file_date_fmt : str
        Format of the date in the dataset.


    Returns
    -------
    dict
        key is a datetime in the time dataset of netcdf file.
        Value is a list of Paths on that date.
        E.g.:
        {(datetime.datetime(2017, 10, 13, 18),
          [Path(...)
            Path(...),
            ...]),
         (datetime.datetime(2017, 10, 25, 18),
          [Path(...)
            Path(...),
            ...]),
        }

    """
    # collapse into groups of dates
    # Use a `defaultdict` so we dont have to sort the files by date in advance,
    # but rather just extend the list each time there's a new group
    grouped_images: dict[datetime.datetime, list[Filename]] = defaultdict(list)

    def get_dates_netcdf(filename: Filename):
        ds = xarray.open_mfdataset(filename)
        date_time = ds.time.values.astype(file_date_fmt).astype(datetime.datetime)
        return date_time

    for date, g in itertools.groupby(files, key=lambda x: tuple(get_dates_netcdf(x))):
        grouped_images[date[0]].extend(list(g))
    return grouped_images


def convert_netcdf(
    weather_model_file,
    lat_lon_bounds,
    weather_model_output_dir,
    weather_model: TropoModel = TropoModel.ECMWF,
):
    """Convert the weather model NetCDF to working NetCDF.

    Parameters.
    ----------
        weather_model_file: str
        The ECMWF/HRES NetCDF weather model file
        lat_lon_bounds: list
            bounding box
        weather_model_output_dir: str
            the output directory of the internal NetCDF file

    Returns
    -------
            the path of the output NetCDF file

    """
    if weather_model == TropoModel.ECMWF:
        from dolphin.atmosphere.weather_model import ECMWF as WMODEL

    elif weather_model == TropoModel.HRES:
        from dolphin.atmosphere.weather_model import HRES as WMODEL
    else:
        raise FileExistsError("Weather model not supported")

    wmodel = WMODEL()

    Path(weather_model_output_dir).mkdir(parents=True, exist_ok=True)
    ds = xarray.open_dataset(weather_model_file)

    # Get the datetime of the weather model file
    weather_model_time = ds.time.values.astype("datetime64[s]").astype(
        datetime.datetime
    )[0]
    ds.close()
    # Set up the time, Lat/Lon, and working location, where
    # the lat/lon bounds are applied to clip the global
    # weather model to minimize the data processing
    wmodel.set_time(weather_model_time)
    wmodel.set_latlon_bounds(ll_bounds=lat_lon_bounds)
    wmodel.set_wm_loc(weather_model_output_dir)

    # Load the ECMWF NetCDF weather model
    wmodel.load_weather(weather_model_file)

    # Process the weather model data
    wmodel._find_e()
    wmodel._uniform_in_z(_zlevels=None)

    # This function fills the NaNs with 0
    wmodel._check_for_nans()

    wmodel._get_wet_refractivity()
    wmodel._get_hydro_refractivity()
    wmodel._adjust_grid(wmodel.get_latlon_bounds())

    # Compute Zenith delays at the weather model grid nodes
    wmodel._get_ztd()

    output_file = wmodel.out_file(weather_model_output_dir)
    wmodel._out_name = output_file

    # Return the ouput file if it exists
    if Path(output_file).exists():
        return output_file
    else:
        # Write to hard drive
        return wmodel.write()



import numpy as np
import numpy.testing as npt
import pytest

from dolphin.phase_link._eigenvalues import (
    eigh_largest_stack,
    eigh_smallest_stack,
)
from dolphin.phase_link.simulate import simulate_coh

# Used for matrix size
N = 20


def get_eigvec_phase_difference(a, b):
    """Compare the phase difference of two eigenvectors (or stacks).

    Using this since there may be a relative offset in the phase of
    each starting, depending on the algorithm.
    """
    a_0 = a * a[..., [0]].conj()
    b_0 = b * b[..., [0]].conj()
    return np.angle(a_0 * b_0.conj())


class TestEighStack:
    """Test the stack functionality to compare to scipy eigh."""

    @pytest.fixture
    def coh_stack(self):
        num_rows, num_cols = 6, 7
        out = np.empty((num_rows, num_cols, N, N), dtype=np.complex64)
        for row in range(num_rows):
            for col in range(num_cols):
                out[row, col] = simulate_coh(num_acq=N, add_signal=True)[0]
        return out

    @pytest.fixture
    def expected_largest(self, coh_stack):
        # Compare to numpy
        eig_vals, eig_vecs = np.linalg.eigh(coh_stack)
        assert np.all(eig_vals >= 0)
        expected_eig = eig_vals[:, :, -1]
        expected_evec = eig_vecs[:, :, :, -1]
        assert expected_eig.shape == (6, 7)
        assert expected_evec.shape == (6, 7, N)
        return expected_eig, expected_evec

    def test_eigh_largest_stack(self, coh_stack, expected_largest):
        expected_eig, expected_evec = expected_largest
        evalues, evecs = eigh_largest_stack(coh_stack)
        assert evalues.shape == (6, 7)
        assert evecs.shape == (6, 7, N)

        npt.assert_allclose(expected_eig, evalues, atol=2e-5)
        assert np.max(np.abs(get_eigvec_phase_difference(expected_evec, evecs))) < 5e-3

    @pytest.fixture
    def coh_gamma_inv_stack(self, coh_stack) -> np.ndarray:
        return coh_stack * np.abs(np.linalg.inv(coh_stack))

    @pytest.fixture
    def expected_smallest(self, coh_gamma_inv_stack):
        # Compare to numpy
        eig_vals, eig_vecs = np.linalg.eigh(coh_gamma_inv_stack)
        assert np.all(eig_vals >= 0)
        expected_eig = eig_vals[:, :, 0]
        expected_evec = eig_vecs[:, :, :, 0]
        assert expected_eig.shape == (6, 7)
        assert expected_evec.shape == (6, 7, N)
        return expected_eig, expected_evec

    def test_eigh_smallest_stack(self, coh_gamma_inv_stack, expected_smallest):
        expected_eig, expected_evec = expected_smallest
        mu = 0.99
        evalues, evecs = eigh_smallest_stack(coh_gamma_inv_stack, mu)
        assert evalues.shape == (6, 7)
        assert evecs.shape == (6, 7, N)

        # Compare the values:
        npt.assert_allclose(evalues, expected_eig, atol=2e-5)

        # Check the max phase difference
        assert np.max(np.abs(get_eigvec_phase_difference(expected_evec, evecs))) < 1e-4



from enum import Enum

__all__ = [
    "ShpMethod",
    "UnwrapMethod",
    "CallFunc",
]


class ShpMethod(str, Enum):
    """Method for finding SHPs during phase linking."""

    GLRT = "glrt"
    KS = "ks"
    RECT = "rect"
    # Alias for no SHP search
    NONE = "rect"


class UnwrapMethod(str, Enum):
    """Phase unwrapping method."""

    SNAPHU = "snaphu"
    ICU = "icu"
    PHASS = "phass"
    SPURT = "spurt"
    WHIRLWIND = "whirlwind"


class CallFunc(str, Enum):
    """Call function for the timeseries method to find reference point."""

    MIN = "min"
    MAX = "max"



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

g0 = np.float64(9.80665)  # Standard gravitational constant
G1 = np.float64(
    9.80616
)  # Gravitational constant @ 45° latitude used for corrections of
# earth's centrifugal force
R_EARTH_MAX_WGS84 = Rmax = 6378137
R_EARTH_MIN_WGS84 = Rmin = 6356752


def find_svp(t):
    """Calculate standard vapor presure. Should be model-specific."""
    # From TRAIN:
    # Could not find the wrf used equation as they appear to be
    # mixed with latent heat etc. Istead I used the equations used
    # in ERA-I (see IFS documentation part 2: Data assimilation
    # (CY25R1)). Calculate saturated water vapour pressure (svp) for
    # water (svpw) using Buck 1881 and for ice (swpi) from Alduchow
    # and Eskridge (1996) euation AERKi

    # TODO: figure out the sources of all these magic numbers and move
    # them somewhere more visible.
    # TODO: (Jeremy) - Need to fix/get the equation for the other
    # weather model types. Right now this will be used for all models,
    # except WRF, which is yet to be implemented in my new structure.
    t1 = 273.15  # O Celsius
    t2 = 250.15  # -23 Celsius

    tref = t - t1
    wgt = (t - t2) / (t1 - t2)
    svpw = 6.1121 * np.exp((17.502 * tref) / (240.97 + tref))
    svpi = 6.1121 * np.exp((22.587 * tref) / (273.86 + tref))

    svp = svpi + (svpw - svpi) * wgt**2
    ix_bound1 = t > t1
    svp[ix_bound1] = svpw[ix_bound1]
    ix_bound2 = t < t2
    svp[ix_bound2] = svpi[ix_bound2]

    svp = svp * 100
    return svp.astype(np.float32)


def calcgeoh(
    lnsp: np.ndarray,
    t: np.ndarray,
    q: np.ndarray,
    z: np.ndarray,
    a: np.ndarray,
    b: np.ndarray,
    R_d: int,
    num_levels: int,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate pressure, geopotential, and geopotential height.

    from the surface pressure and model levels provided by a weather model.
    The model levels are numbered from the highest elevation to the lowest.

    Parameters
    ----------
    lnsp: ndarray
        array of log surface pressure
    t: ndarray
        cube of temperatures
    q: ndarray
        cube of specific humidity
    z: ndarray
        cube of surface heights
    a: ndarray
        vector of a values
    b: ndarray
        vector of b values
    R_d: int
        model parameter
    num_levels: int
        integer number of model levels

    Returns
    -------
    geopotential : np.ndarray
        The geopotential in units of height times acceleration
    pressurelvs : np.ndarray
        The pressure at each of the model levels for each of
                    the input points
    geoheight : np.ndarray
        The geopotential heights

    """
    geopotential = np.zeros_like(t)
    pressurelvs = np.zeros_like(geopotential)
    geoheight = np.zeros_like(geopotential)

    # log surface pressure
    # Note that we integrate from the ground up, so from the largest model level to 0
    sp = np.exp(lnsp)

    if len(a) != num_levels + 1 or len(b) != num_levels + 1:
        raise ValueError(
            "I have here a model with {} levels, but parameters a ".format(num_levels)
            + "and b have lengths {} and {} respectively. Of ".format(len(a), len(b))
            + "course, these three numbers should be equal."
        )

    # Integrate up into the atmosphere from *lowest level*
    z_h = 0  # initial value
    for lev, t_level, q_level in zip(range(num_levels, 0, -1), t[::-1], q[::-1]):
        # lev is the level number 1-60, we need a corresponding index
        # into ts and qs
        # ilevel = num_levels - lev # << this was Ray's original, but is a typo
        # because indexing like that results in pressure and height arrays that
        # are in the opposite orientation to the t/q arrays.
        ilevel = lev - 1

        # compute moist temperature
        t_lev = t_level * (1 + 0.609133 * q_level)

        # compute the pressures (on half-levels)
        Ph_lev = a[lev - 1] + (b[lev - 1] * sp)
        Ph_levplusone = a[lev] + (b[lev] * sp)

        pressurelvs[ilevel] = (
            Ph_lev  # + Ph_levplusone) / 2
            # average pressure at half-levels above and below
        )

        if lev == 1:
            dlogP = np.log(Ph_levplusone / 0.1)
            alpha = np.log(2)
        else:
            dlogP = np.log(Ph_levplusone) - np.log(Ph_lev)
            alpha = 1 - ((Ph_lev / (Ph_levplusone - Ph_lev)) * dlogP)

        TRd = t_lev * R_d

        # z_f is the geopotential of this full level
        # integrate from previous (lower) half-level z_h to the full level
        z_f = z_h + TRd * alpha + z

        # Geopotential (add in surface geopotential)
        geopotential[ilevel] = z_f
        geoheight[ilevel] = geopotential[ilevel] / g0

        # z_h is the geopotential of 'half-levels'
        # integrate z_h to next half level
        z_h += TRd * dlogP

    return geopotential, pressurelvs, geoheight


def geo_to_ht(lats: np.ndarray, hts: np.ndarray) -> np.ndarray:
    """Convert geopotential height to ellipsoidal heights.

    referenced to WGS84.
    Note that this formula technically computes height above
    geoid (geometric height) but the geoid is actually a
    perfect sphere; Thus returned heights are above a reference
    ellipsoid, which most assume to be a sphere.
    However, by calculating the ellipsoid here we directly
    reference to WGS84.

    # h = (geopotential * Re) / (g0 * Re - geopotential)
    # Assumes a sphere instead of an ellipsoid

    Parameters
    ----------
    lats : np.ndarray
        latitude of points of interest
    hts : np.ndarray
        geopotential height at points of interest

    Returns
    -------
    ndarray: geometric heights.
        These are approximate ellipsoidal heights referenced to WGS84

    """
    g_ll = _get_g_ll(lats)  # gravity function of latitude
    Re = get_earth_radius(lats)  # Earth radius function of latitude

    # Calculate Geometric Height, h
    h = (hts * Re) / (g_ll / g0 * Re - hts)

    return h


def _get_g_ll(lats):
    """Compute the variation in gravity constant with latitude."""
    return G1 * (1 - 0.002637 * cosd(2 * lats) + 0.0000059 * (cosd(2 * lats)) ** 2)


def cosd(x):
    """Return the cosine of x when x is in degrees."""
    return np.cos(np.radians(x))


def get_earth_radius(lats: np.ndarray) -> np.ndarray:
    """Get the earth radius as a function of latitude for WGS84.

    Parameters
    ----------
    lats : np.ndarray
        ndarray of geodetic latitudes in degrees

    Returns
    -------
    np.ndarray
        ndarray of earth radius at each latitude

    Example
    -------
    >>> import numpy as np
    >>> output = get_earth_radius(np.array([0, 30, 45, 60, 90]))
    >>> output
    array([6378137. , 6372770.5219805 , 6367417.56705189, 6362078.07851428, 6356752. ])
    >>> assert output[0] == 6378137  # (Rmax)
    >>> assert output[-1] == 6356752  # (Rmin)

    """
    return np.sqrt(1 / (((cosd(lats) ** 2) / Rmax**2) + ((sind(lats) ** 2) / Rmin**2)))


def sind(x):
    """Return the sine of x when x is in degrees."""
    return np.sin(np.radians(x))


def pad_lower(invar):
    """Add a layer of data below the lowest current z-level at height zmin."""
    new_var = _least_nonzero(invar)
    return np.concatenate((new_var[:, :, np.newaxis], invar), axis=2)


def _least_nonzero(a):
    """Fill in a flat array with the first non-nan value in the last dimension.

    Useful for interpolation below the bottom of the weather model.
    """
    mgrid_index = tuple(slice(None, d) for d in a.shape[:-1])
    return a[(*tuple(np.mgrid[mgrid_index]), (~np.isnan(a)).argmax(-1))]


def interpolate_along_axis(oldCoord, newCoord, data, axis=2):
    """Interpolate an array of 3-D data along one axis. This function.

    assumes that the x-coordinate increases monotonically.
    """
    if oldCoord.ndim > 1:
        stackedData = np.concatenate([oldCoord, data, newCoord], axis=axis)
        out = np.apply_along_axis(
            interp_vector, axis=axis, arr=stackedData, Nx=oldCoord.shape[axis]
        )
    else:
        out = np.apply_along_axis(
            interp_v,
            axis=axis,
            arr=data,
            old_x=oldCoord,
            new_x=newCoord,
            left=np.nan,
            right=np.nan,
        )

    return out


def interp_vector(vec, Nx):
    """Interpolate data from a single vector containing the original.

    x, the original y, and the new x, in that order. Nx tells the
    number of original x-points.
    """
    x = vec[:Nx]
    y = vec[Nx : 2 * Nx]
    xnew = vec[2 * Nx :]
    f = interp1d(x, y, bounds_error=False, copy=False, assume_sorted=True)
    return f(xnew)


def interp_v(y, old_x, new_x, left=None, right=None, period=None):
    """Rearrange np.interp's arguments."""
    return np.interp(new_x, old_x, y, left=left, right=right, period=period)


def fillna3d(array: np.ndarray, axis: int = -1, fill_value: float = 0.0) -> np.ndarray:
    """Fill in NaNs in 3D arrays using the nearest non-NaN value for "low" NaNs.

    and a specified fill value (default 0.0) for "high" NaNs.

    Parameters
    ----------
    array : np.ndarray
        3D array, where the last axis is the "z" dimension.
    axis : int, optional
        The axis along which to fill values. Default is -1 (the last axis).
    fill_value : float, optional
        The value used for filling NaNs. Default is 0.0.

    Returns
    -------
    np.ndarray
        3D array with low NaNs filled using nearest neighbors and high NaNs
        filled with the specified fill value.

    """
    # fill lower NaNs with nearest neighbor
    narr = np.moveaxis(array, axis, -1)
    nars = narr.reshape((np.prod(narr.shape[:-1]), narr.shape[-1]))
    dfd = pd.DataFrame(data=nars).interpolate(axis=1, limit_direction="backward")
    out = dfd.values.reshape(array.shape)

    # fill upper NaNs with 0s
    outmat = np.moveaxis(out, -1, axis)
    outmat[np.isnan(outmat)] = fill_value
    return outmat



"""Estimate wrapped phase for one ministack of SLCs."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np
from numpy.typing import DTypeLike
from tqdm.auto import tqdm

from dolphin import io, shp
from dolphin._decorators import atomic_output
from dolphin._types import Filename, HalfWindow, Strides
from dolphin.io import EagerLoader, StridedBlockManager, VRTStack
from dolphin.masking import load_mask_as_numpy
from dolphin.phase_link import PhaseLinkRuntimeError, compress, run_phase_linking
from dolphin.ps import calc_ps_block
from dolphin.stack import MiniStackInfo

from .config import ShpMethod

logger = logging.getLogger(__name__)

__all__ = ["run_wrapped_phase_single"]


@dataclass
class OutputFile:
    filename: Path
    dtype: DTypeLike
    strides: Optional[dict[str, int]] = None
    nbands: int = 1


@atomic_output(output_arg="output_folder", is_dir=True)
def run_wrapped_phase_single(
    *,
    slc_vrt_file: Filename,
    ministack: MiniStackInfo,
    output_folder: Filename,
    half_window: dict,
    strides: Optional[dict] = None,
    reference_idx: int = 0,
    beta: float = 0.00,
    use_evd: bool = False,
    mask_file: Optional[Filename] = None,
    ps_mask_file: Optional[Filename] = None,
    amp_mean_file: Optional[Filename] = None,
    amp_dispersion_file: Optional[Filename] = None,
    shp_method: ShpMethod = ShpMethod.NONE,
    shp_alpha: float = 0.05,
    shp_nslc: Optional[int] = None,
    block_shape: tuple[int, int] = (1024, 1024),
    baseline_lag: Optional[int] = None,
    **tqdm_kwargs,
):
    """Estimate wrapped phase for one ministack.

    Output files will all be placed in the provided `output_folder`.
    """
    # TODO: extract common stuff between here and sequential
    if strides is None:
        strides = {"x": 1, "y": 1}
    strides_tup = Strides(y=strides["y"], x=strides["x"])
    half_window_tup = HalfWindow(y=half_window["y"], x=half_window["x"])
    output_folder = Path(output_folder)
    vrt = VRTStack.from_vrt_file(slc_vrt_file)
    input_slc_files = ministack.file_list
    assert len(input_slc_files) == vrt.shape[0]

    # If we are using a different number of SLCs for the amplitude data,
    # we should note that for the SHP finding algorithms
    if shp_nslc is None:
        shp_nslc = len(input_slc_files)

    logger.info(f"{vrt}: from {ministack.dates[0]} {ministack.file_list[-1]}")

    nrows, ncols = vrt.shape[-2:]

    nodata_mask = _get_nodata_mask(mask_file, nrows, ncols)
    ps_mask = _get_ps_mask(ps_mask_file, nrows, ncols)
    amp_mean, amp_variance = _get_amp_mean_variance(amp_mean_file, amp_dispersion_file)
    amp_stack: Optional[np.ndarray] = None

    xhalf, yhalf = half_window["x"], half_window["y"]

    # If we were passed any compressed SLCs in `input_slc_files`,
    # then we want that index for when we create new compressed SLCs.
    # We skip the old compressed SLCs to create new ones
    first_real_slc_idx = ministack.first_real_slc_idx

    msg = (
        f"Processing {len(input_slc_files) - first_real_slc_idx} SLCs +"
        f" {first_real_slc_idx} compressed SLCs. "
    )
    logger.info(msg)

    # Create the background writer for this ministack
    writer = io.BackgroundBlockWriter()

    logger.info(f"Total stack size (in pixels): {vrt.shape}")
    # Set up the output folder with empty files to write into
    phase_linked_slc_files = setup_output_folder(
        ministack=ministack,
        driver="GTiff",
        strides=strides,
        output_folder=output_folder,
        like_filename=vrt.outfile,
        nodata=0,
    )

    comp_slc_info = ministack.get_compressed_slc_info()

    # Use the real-SLC date range for output file naming
    start_end = ministack.real_slc_date_range_str
    output_files: list[OutputFile] = [
        # The compressed SLC does not used strides, but has extra band for dispersion
        OutputFile(output_folder / comp_slc_info.filename, np.complex64, nbands=2),
        # but all the rest do:
        OutputFile(
            output_folder / f"temporal_coherence_{start_end}.tif", np.float32, strides
        ),
        OutputFile(output_folder / f"shp_counts_{start_end}.tif", np.uint16, strides),
        OutputFile(output_folder / f"eigenvalues_{start_end}.tif", np.float32, strides),
        OutputFile(output_folder / f"estimator_{start_end}.tif", np.int8, strides),
        OutputFile(output_folder / f"avg_coh_{start_end}.tif", np.uint16, strides),
    ]
    for op in output_files:
        io.write_arr(
            arr=None,
            like_filename=vrt.outfile,
            output_name=op.filename,
            dtype=op.dtype,
            strides=op.strides,
            nbands=op.nbands,
            nodata=0,
        )

    # Iterate over the output grid
    block_manager = StridedBlockManager(
        arr_shape=(nrows, ncols),
        block_shape=block_shape,
        strides=strides_tup,
        half_window=half_window_tup,
    )
    # Set up the background loader
    loader = EagerLoader(reader=vrt, block_shape=block_shape)
    # Queue all input slices, skip ones that are all nodata
    blocks = []
    # Queue all input slices, skip ones that are all nodata
    for b in block_manager.iter_blocks():
        in_rows, in_cols = b[2]
        if nodata_mask[in_rows, in_cols].all():
            continue
        loader.queue_read(in_rows, in_cols)
        blocks.append(b)

    logger.info(f"Iterating over {block_shape} blocks, {len(blocks)} total")
    for (
        (out_rows, out_cols),
        (out_trim_rows, out_trim_cols),
        (in_rows, in_cols),
        (in_no_pad_rows, in_no_pad_cols),
        (in_trim_rows, in_trim_cols),
    ) in tqdm(blocks, **tqdm_kwargs):
        logger.debug(f"{out_rows = }, {out_cols = }, {in_rows = }, {in_no_pad_rows = }")

        cur_data, (read_rows, read_cols) = loader.get_data()
        if np.all(cur_data == 0) or np.isnan(cur_data).all():
            continue
        assert read_rows == in_rows and read_cols == in_cols

        cur_data = cur_data.astype(np.complex64)

        if shp_method == "ks":
            # Only actually compute if we need this one
            amp_stack = np.abs(cur_data)

        # Compute the neighbor_arrays for this block
        neighbor_arrays = shp.estimate_neighbors(
            halfwin_rowcol=(yhalf, xhalf),
            alpha=shp_alpha,
            strides=strides,
            mean=amp_mean[in_rows, in_cols] if amp_mean is not None else None,
            var=amp_variance[in_rows, in_cols] if amp_variance is not None else None,
            nslc=shp_nslc,
            amp_stack=amp_stack,
            method=shp_method,
        )
        # Run the phase linking process on the current ministack
        reference_idx = max(0, first_real_slc_idx - 1)
        try:
            pl_output = run_phase_linking(
                cur_data,
                half_window=half_window_tup,
                strides=strides_tup,
                use_evd=use_evd,
                beta=beta,
                reference_idx=reference_idx,
                nodata_mask=nodata_mask[in_rows, in_cols],
                ps_mask=ps_mask[in_rows, in_cols],
                neighbor_arrays=neighbor_arrays,
                baseline_lag=baseline_lag,
                avg_mag=amp_mean[in_rows, in_cols] if amp_mean is not None else None,
            )
        except PhaseLinkRuntimeError as e:
            # note: this is a warning instead of info, since it should
            # get caught at the "skip_empty" step
            msg = f"At block {in_rows.start}, {in_cols.start}: {e}"
            if "are all NaNs" in e.args[0]:
                # Some SLCs in the ministack are all NaNs
                # This happens from a shifting burst window near the edges,
                # and seems to cause no issues
                logger.debug(msg)
            else:
                logger.warning(msg)
            continue

        # Fill in the nan values with 0
        np.nan_to_num(pl_output.cpx_phase, copy=False)
        np.nan_to_num(pl_output.temp_coh, copy=False)

        # Save each of the MLE estimates (ignoring those corresponding to
        # compressed SLCs indexes)
        assert len(pl_output.cpx_phase[first_real_slc_idx:]) == len(
            phase_linked_slc_files
        )

        for img, f in zip(
            pl_output.cpx_phase[first_real_slc_idx:, out_trim_rows, out_trim_cols],
            phase_linked_slc_files,
        ):
            writer.queue_write(img, f, out_rows.start, out_cols.start)

        # Compress the ministack using only the non-compressed SLCs
        # Get the mean to set as pixel magnitudes
        abs_stack = np.abs(cur_data[first_real_slc_idx:, in_trim_rows, in_trim_cols])
        cur_data_mean, cur_amp_dispersion, _ = calc_ps_block(abs_stack)
        cur_comp_slc = compress(
            # Get the inner portion of the full-res SLC data
            cur_data[first_real_slc_idx:, in_trim_rows, in_trim_cols],
            pl_output.cpx_phase[first_real_slc_idx:, out_trim_rows, out_trim_cols],
            slc_mean=cur_data_mean,
        )
        # TODO: truncate

        # ### Save results ###

        # Save the compressed SLC block
        writer.queue_write(
            cur_comp_slc,
            output_files[0].filename,
            in_no_pad_rows.start,
            in_no_pad_cols.start,
            band=1,
        )
        # Save the amplitude dispersion of the real SLC data
        writer.queue_write(
            cur_amp_dispersion,
            output_files[0].filename,
            in_no_pad_rows.start,
            in_no_pad_cols.start,
            band=2,
        )

        # All other outputs are strided (smaller in size)
        out_datas = [
            pl_output.temp_coh,
            pl_output.shp_counts,
            pl_output.eigenvalues,
            pl_output.estimator,
            pl_output.avg_coh,
        ]
        for data, output_file in zip(out_datas, output_files[1:]):
            if data is None:  # May choose to skip some outputs, e.g. "avg_coh"
                continue
            writer.queue_write(
                data[out_trim_rows, out_trim_cols],
                output_file.filename,
                out_rows.start,
                out_cols.start,
            )

    loader.notify_finished()
    # Block until all the writers for this ministack have finished
    logger.info(f"Waiting to write {writer.num_queued} blocks of data.")
    writer.notify_finished()
    logger.info(f"Finished ministack of size {vrt.shape}.")

    logger.info("Repacking for more compression")
    io.repack_rasters(phase_linked_slc_files, keep_bits=12)

    written_comp_slc = output_files[0]

    io.repack_raster(written_comp_slc.filename, keep_bits=12)
    ccslc_info = ministack.get_compressed_slc_info()
    ccslc_info.write_metadata(output_file=written_comp_slc.filename)
    # TODO: Does it make sense to return anything from this?
    # or just allow user to search through the `output_folder` they provided?


def _get_nodata_mask(
    mask_file: Optional[Filename],
    nrows: int,
    ncols: int,
) -> np.ndarray:
    if mask_file is not None:
        return load_mask_as_numpy(mask_file)
    else:
        return np.zeros((nrows, ncols), dtype=bool)


def _get_ps_mask(
    ps_mask_file: Optional[Filename], nrows: int, ncols: int
) -> np.ndarray:
    if ps_mask_file is not None:
        ps_mask = io.load_gdal(ps_mask_file, masked=True)
        # Fill the nodata values with false
        ps_mask = ps_mask.astype(bool).filled(False)
    else:
        ps_mask = np.zeros((nrows, ncols), dtype=bool)
    return ps_mask


def _get_amp_mean_variance(
    amp_mean_file: Optional[Filename],
    amp_dispersion_file: Optional[Filename],
) -> tuple[np.ndarray, np.ndarray]:
    if amp_mean_file is not None and amp_dispersion_file is not None:
        # Note: have to fill, since numba (as of 0.57) can't do masked arrays
        amp_mean = io.load_gdal(amp_mean_file, masked=True).filled(np.nan)
        amp_dispersion = io.load_gdal(amp_dispersion_file, masked=True).filled(np.nan)
        # convert back to variance from dispersion: amp_disp = std_dev / mean
        amp_variance = (amp_dispersion * amp_mean) ** 2
    else:
        amp_mean = amp_variance = None

    return amp_mean, amp_variance


def setup_output_folder(
    ministack: MiniStackInfo,
    driver: str = "GTiff",
    dtype="complex64",
    like_filename: Optional[Filename] = None,
    strides: Optional[dict[str, int]] = None,
    nodata: Optional[float] = 0,
    output_folder: Optional[Path] = None,
) -> list[Path]:
    """Create empty output files for each band after `start_idx` in `vrt_stack`.

    Also creates an empty file for the compressed SLC.
    Used to prepare output for block processing.

    Parameters
    ----------
    ministack : MiniStackInfo
        [dolphin.stack.MiniStackInfo][] object for the current batch of SLCs
    driver : str, optional
        Name of GDAL driver, by default "GTiff"
    dtype : str, optional
        Numpy datatype of output files, by default "complex64"
    like_filename : Filename, optional
        Filename to use for getting the shape/GDAL metadata of the output files.
        If None, will use the first SLC in `vrt_stack`
    strides : dict[str, int], optional
        Strides to use when creating the empty files, by default {"y": 1, "x": 1}
        Larger strides will create smaller output files, computed using
        [dolphin.io.compute_out_shape][]
    nodata : float, optional
        Nodata value to use for the output files, by default 0.
    output_folder : Path, optional
        Path to output folder, by default None
        If None, will use the same folder as the first SLC in `vrt_stack`

    Returns
    -------
    list[Path]
        list of saved empty files for the outputs of phase linking

    """
    if strides is None:
        strides = {"y": 1, "x": 1}
    if output_folder is None:
        output_folder = ministack.output_folder
    # Note: DONT use the ministack.output_folder here, since that will
    # be the tempdir made by @atomic_output
    # # output_folder = Path(ministack.output_folder)
    output_folder.mkdir(exist_ok=True, parents=True)

    start_idx = ministack.first_real_slc_idx
    date_strs = ministack.get_date_str_list()[start_idx:]

    phase_linked_slc_files = []
    for filename in date_strs:
        slc_name = Path(filename).stem
        output_path = output_folder / f"{slc_name}.slc.tif"

        io.write_arr(
            arr=None,
            like_filename=like_filename,
            output_name=output_path,
            driver=driver,
            nbands=1,
            dtype=dtype,
            strides=strides,
            nodata=nodata,
        )

        phase_linked_slc_files.append(output_path)
    return phase_linked_slc_files



from __future__ import annotations

import logging
import warnings

import numpy as np
from numba import njit
from numpy.typing import ArrayLike

from dolphin._types import Strides
from dolphin.utils import take_looks

logger = logging.getLogger(__name__)


def fill_ps_pixels(
    cpx_phase: ArrayLike,
    temp_coh: ArrayLike,
    slc_stack: ArrayLike,
    ps_mask: ArrayLike,
    strides: Strides,
    avg_mag: ArrayLike | None,
    reference_idx: int = 0,
    use_max_ps: bool = True,
):
    """Fill in the PS locations in the MLE estimate with the original SLC data.

    Overwrites `cpx_phase` and `temp_coh` in place.

    Parameters
    ----------
    cpx_phase : ndarray, shape = (nslc, rows, cols)
        The complex valued-MLE estimate of the phase.
    temp_coh : ndarray, shape = (rows, cols)
        The temporal coherence of the estimate.
    slc_stack : ArrayLike
        The original SLC stack, with shape (n_images, n_rows, n_cols)
    ps_mask : ndarray, shape = (rows, cols)
        Boolean mask of pixels marking persistent scatterers (PS).
    strides : Strides
        The decimation (y, x) factors
    avg_mag : ArrayLike, optional
        The average magnitude of the SLC stack, used to to find the brightest
        PS pixels to fill within each look window.
    reference_idx : int, default = 0
        SLC to use as reference for PS pixels. All pixel values are multiplied
        by the conjugate of this index
    use_max_ps : bool, optional, default = True
        If True, use the brightest PS pixel in each look window to fill in the
        MLE estimate. If False, use the average of all PS pixels in each look window.

    """
    if avg_mag is None:
        # Get the average magnitude of the SLC stack
        # nanmean will ignore single NaNs, but not all NaNs, per pixel
        with warnings.catch_warnings():
            # ignore the warning about nansum/nanmean of empty slice
            warnings.simplefilter("ignore", category=RuntimeWarning)
            avg_mag = np.nanmean(np.abs(slc_stack), axis=0)
    mag = avg_mag.copy()

    # null out all the non-PS pixels when finding the brightest PS pixels
    mag[~ps_mask] = np.nan
    # For ps_mask, we set to True if any pixels within the window were PS
    ps_mask_looked = take_looks(ps_mask, *strides, func_type="any", edge_strategy="pad")
    # make sure it's the same size as the MLE result/temp_coh after padding
    ps_mask_looked = ps_mask_looked[: cpx_phase.shape[1], : cpx_phase.shape[2]]

    if use_max_ps:
        logger.debug("Using max PS pixel to fill in MLE estimate")
        # Get the indices of the brightest pixels within each look window
        slc_r_idxs, slc_c_idxs = get_max_idxs(mag, *strides)

        # we're only filling where there are PS pixels
        ref = np.exp(-1j * np.angle(slc_stack[reference_idx][slc_r_idxs, slc_c_idxs]))
        for i in range(len(slc_stack)):
            slc_phase = np.angle(slc_stack[i][slc_r_idxs, slc_c_idxs])
            cur_amp = np.abs(cpx_phase[i][ps_mask_looked])
            new_value = np.exp(1j * slc_phase) * ref
            cpx_phase[i][ps_mask_looked] = cur_amp * new_value

    else:
        # Get the average of all PS pixels within each look window
        # The referencing to SLC 0 is done in _get_avg_ps
        avg_ps = _get_avg_ps(slc_stack, ps_mask, strides)[
            :, : cpx_phase.shape[1], : cpx_phase.shape[2]
        ]
        ps_phases = np.angle(avg_ps[:, ps_mask_looked])

        # Set the angle only, don't change magnitude
        cpx_phase[:, ps_mask_looked] = np.abs(cpx_phase[:, ps_mask_looked]) * np.exp(
            1j * ps_phases
        )

    # Force PS pixels to have high temporal coherence
    temp_coh[ps_mask_looked] = 1


def _get_avg_ps(
    slc_stack: np.ndarray, ps_mask: np.ndarray, strides: Strides
) -> np.ndarray:
    # First, set all non-PS pixels to NaN
    slc_stack_nanned = slc_stack.copy()
    slc_stack_nanned[:, ~ps_mask] = np.nan
    # Reference all ps pixels in the SLC stack to the first SLC
    slc_stack_nanned[:, ps_mask] *= np.exp(
        -1j * np.angle(slc_stack_nanned[0, ps_mask])
    )[None]
    # Then, take the average of all PS pixels within each look window
    return take_looks(
        slc_stack_nanned,
        *strides,
        func_type="nanmean",
        edge_strategy="pad",
    )


@njit
def get_max_idxs(arr: ArrayLike, row_looks: int, col_looks: int):
    window_height, window_width = row_looks, col_looks
    height, width = arr.shape

    # Calculate the number of windows
    new_height = height // window_height
    new_width = width // window_width

    row_indices = []
    col_indices = []

    for i in range(new_height):
        for j in range(new_width):
            window = arr[
                i * window_height : (i + 1) * window_height,
                j * window_width : (j + 1) * window_width,
            ]

            max_value = -np.inf
            max_row, max_col = -1, -1

            for row in range(window_height):
                for col in range(window_width):
                    value = window[row, col]
                    if not np.isnan(value) and value > max_value:
                        max_value = value
                        max_row = row
                        max_col = col

            # If max_row and max_col are still -1, it means the window was all NaNs
            if max_row != -1 and max_col != -1:
                row_indices.append(i * window_height + max_row)
                col_indices.append(j * window_width + max_col)

    return np.array(row_indices), np.array(col_indices)



import argparse
import sys

import dolphin._cli_filter
import dolphin._cli_timeseries
import dolphin._cli_unwrap
import dolphin.workflows._cli_config
import dolphin.workflows._cli_run
from dolphin import __version__


def main(args=None):
    """Top-level command line interface to the workflows."""
    parser = argparse.ArgumentParser(
        prog=__package__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparser = parser.add_subparsers(title="subcommands")
    parser.add_argument("--version", action="version", version=__version__)

    # Adds the subcommand to the top-level parser
    dolphin.workflows._cli_run.get_parser(subparser, "run")
    dolphin.workflows._cli_config.get_parser(subparser, "config")
    dolphin._cli_unwrap.get_parser(subparser, "unwrap")
    dolphin._cli_timeseries.get_parser(subparser, "timeseries")
    dolphin._cli_filter.get_parser(subparser, "filter")
    parsed_args = parser.parse_args(args=args)

    arg_dict = vars(parsed_args)
    try:
        run_func = arg_dict.pop("run_func")
        run_func(**arg_dict)
    except KeyError:
        # No arguments passed
        parser.print_help()
        sys.exit(1)



from __future__ import annotations

import os
from pathlib import Path

import pytest

from dolphin.io import get_raster_xysize
from dolphin.workflows import config, ps


def test_ps_workflow_run(opera_slc_files_official: list[Path], tmpdir):
    # get the first half of the files, for just one burst stack
    file_list = opera_slc_files_official[: len(opera_slc_files_official) // 2]
    with tmpdir.as_cwd():
        cfg = config.PsWorkflow(
            cslc_file_list=file_list,
            input_options={"subdataset": "/data/VV"},
            worker_settings={
                "gpu_enabled": (os.environ.get("NUMBA_DISABLE_JIT") != "1")
            },
        )
        ps.run(cfg)

        work_dir = cfg.work_directory
        # Check the output files were made
        expected_outputs = [
            work_dir / "PS" / "ps_pixels.tif",
            work_dir / "PS" / "amp_mean.tif",
            work_dir / "PS" / "amp_dispersion.tif",
        ]
        for f in expected_outputs:
            assert f.exists()

        in_size = get_raster_xysize(f"NETCDF:{file_list[0]}:/data/VV")
        out_size = get_raster_xysize(expected_outputs[0])
        assert in_size == out_size


def test_ps_workflow_multi_burst(opera_slc_files_official: list[Path], tmpdir):
    with tmpdir.as_cwd():
        cfg = config.PsWorkflow(
            cslc_file_list=opera_slc_files_official,
            input_options={"subdataset": "/data/VV"},
            worker_settings={
                "gpu_enabled": (os.environ.get("NUMBA_DISABLE_JIT") != "1")
            },
            log_file=Path() / "dolphin_ps.log",
        )
        with pytest.raises(NotImplementedError):
            ps.run(cfg)



from __future__ import annotations

from contextlib import AbstractContextManager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    Protocol,
    Sequence,
    TypeVar,
    runtime_checkable,
)

import numpy as np
import rasterio
import rasterio.errors
from numpy.typing import ArrayLike, DTypeLike
from rasterio.windows import Window

from dolphin._types import Filename

from ._background import BackgroundWriter
from ._utils import _unpack_3d_slices

__all__ = [
    "BackgroundBlockWriter",
    "DatasetWriter",
    "DatasetStackWriter",
    "RasterWriter",
    "BackgroundRasterWriter",
    "BackgroundStackWriter",
]

if TYPE_CHECKING:
    from dolphin._types import Index


class BackgroundBlockWriter(BackgroundWriter):
    """Class to write data to multiple files in the background using `gdal` bindings."""

    def __init__(self, *, max_queue: int = 0, debug: bool = False, **kwargs):
        super().__init__(nq=max_queue, name="Writer")
        if debug:
            #  background thread. Just synchronously write data
            self.notify_finished()
            self.queue_write = self.write  # type: ignore[assignment]

    def write(
        self,
        data: ArrayLike,
        filename: Filename,
        row_start: int,
        col_start: int,
        band: int | None = None,
    ):
        """Write out an ndarray to a subset of the pre-made `filename`.

        Parameters
        ----------
        data : ArrayLike
            2D or 3D data array to save.
        filename : Filename
            list of output files to save to, or (if cur_block is 2D) a single file.
        row_start : int
            Row index to start writing at.
        col_start : int
            Column index to start writing at.
        band : int, optional
            Band index in the file to write. Defaults to None, which uses first band,
            or whole raster for 3D data.

        Raises
        ------
        ValueError
            If length of `output_files` does not match length of `cur_block`.

        """
        from dolphin.io import write_block

        write_block(data, filename, row_start, col_start, band=band)


@runtime_checkable
class DatasetWriter(Protocol):
    """An array-like interface for writing output datasets.

    `DatasetWriter` defines the abstract interface that types must conform to in order
    to be used by functions which write outputs in blocks.
    Such objects must export NumPy-like `dtype`, `shape`, and `ndim` attributes,
    and must support NumPy-style slice-based indexing for setting data..
    """

    dtype: np.dtype
    """numpy.dtype : Data-type of the array's elements."""

    shape: tuple[int, ...]
    """tuple of int : Tuple of array dimensions."""

    ndim: int
    """int : Number of array dimensions."""

    def __setitem__(self, key: tuple[Index, ...], value: np.ndarray, /) -> None:
        """Write a block of data."""
        ...


@runtime_checkable
class DatasetStackWriter(Protocol):
    """An array-like interface for writing output datasets.

    `DatasetWriter` defines the abstract interface that types must conform to in order
    to be used by functions which write outputs in blocks.
    Such objects must export NumPy-like `dtype`, `shape`, and `ndim` attributes,
    and must support NumPy-style slice-based indexing for setting data..
    """

    dtype: np.dtype
    """numpy.dtype : Data-type of the array's elements."""

    shape: tuple[int, ...]
    """tuple of int : Tuple of array dimensions."""

    ndim: int = 3
    """int : Number of array dimensions."""

    def __setitem__(self, key: tuple[Index, ...], value: np.ndarray, /) -> None:
        """Write a block of data."""
        ...


RasterT = TypeVar("RasterT", bound="RasterWriter")


@dataclass
class RasterWriter(DatasetWriter, AbstractContextManager["RasterWriter"]):
    """A single raster band in a GDAL-compatible dataset containing one or more bands.

    `Raster` provides a convenient interface for using SNAPHU to unwrap ground-projected
    interferograms in raster formats supported by the Geospatial Data Abstraction
    Library (GDAL). It acts as a thin wrapper around a Rasterio dataset and a band
    index, providing NumPy-like access to the underlying raster data.

    Data access is performed lazily -- the raster contents are not stored in memory
    unless/until they are explicitly accessed by an indexing operation.

    `Raster` objects must be closed after use in order to ensure that any data written
    to them is flushed to disk and any associated file objects are closed. The `Raster`
    class implements Python's context manager protocol, which can be used to reliably
    ensure that the raster is closed upon exiting the context manager.
    """

    filename: Filename
    """str or Path : Path to the file to write."""
    band: int = 1
    """int : Band index in the file to write."""

    def __post_init__(self) -> None:
        # Open the dataset.
        self.dataset = rasterio.open(self.filename, mode="r+")

        # Check that `band` is a valid band index in the dataset.
        nbands = self.dataset.count
        if not (1 <= self.band <= nbands):
            errmsg = (
                f"band index {self.band} out of range: dataset contains {nbands} raster"
                " band(s)"
            )
            raise IndexError(errmsg)

        self.ndim = 2

    @classmethod
    def create(
        cls: type[RasterT],
        fp: Filename,
        width: int | None = None,
        height: int | None = None,
        dtype: DTypeLike | None = None,
        driver: str | None = None,
        crs: str | Mapping[str, str] | rasterio.crs.CRS | None = None,
        transform: rasterio.transform.Affine | None = None,
        *,
        like_filename: Filename | None = None,
        **kwargs: Any,
    ) -> RasterT:
        """Create a new single-band raster dataset.

        If another raster is passed via the `like` argument, the new dataset will
        inherit the shape, data-type, driver, coordinate reference system (CRS), and
        geotransform of the reference raster. Driver-specific dataset creation options
        such as chunk size and compression flags may also be inherited.

        All other arguments take precedence over `like` and may be used to override
        attributes of the reference raster when creating the new raster.

        Parameters
        ----------
        fp : str or path-like
            File system path or URL of the local or remote dataset.
        width, height : int or None, optional
            The numbers of columns and rows of the raster dataset. Required if `like` is
            not specified. Otherwise, if None, the new dataset is created with the same
            width/height as `like`. Defaults to None.
        dtype : data-type or None, optional
            Data-type of the raster dataset's elements. Must be convertible to a
            `numpy.dtype` object and must correspond to a valid GDAL datatype. Required
            if `like` is not specified. Otherwise, if None, the new dataset is created
            with the same data-type as `like`. Defaults to None.
        driver : str or None, optional
            Raster format driver name. If None, the method will attempt to infer the
            driver from the file extension. Defaults to None.
        crs : str, dict, rasterio.crs.CRS, or None; optional
            The coordinate reference system. If None, the CRS of `like` will be used, if
            available, otherwise the raster will not be georeferenced. Defaults to None.
        transform : rasterio.transform.Affine or None, optional
            Affine transformation mapping the pixel space to geographic space. If None,
            the geotransform of `like` will be used, if available, otherwise the default
            transform will be used. Defaults to None.
        like_filename : Raster or None, optional
            An optional reference raster. If not None, the new raster will be created
            with the same metadata (shape, data-type, driver, CRS/geotransform, etc) as
            the reference raster. All other arguments will override the corresponding
            attribute of the reference raster. Defaults to None.
        **kwargs : dict, optional
            Additional driver-specific creation options passed to `rasterio.open`.

        """
        if like_filename is not None:
            with rasterio.open(like_filename) as dataset:
                kwargs = dataset.profile | kwargs

        if width is not None:
            kwargs["width"] = width
        if height is not None:
            kwargs["height"] = height
        if dtype is not None:
            kwargs["dtype"] = np.dtype(dtype)
        if driver is not None:
            kwargs["driver"] = driver
        if crs is not None:
            kwargs["crs"] = crs
        if transform is not None:
            kwargs["transform"] = transform

        # Always create a single-band dataset, even if `like` was part of a multi-band
        # dataset.
        kwargs["count"] = 1

        # Create the new single-band dataset.
        with rasterio.open(fp, mode="w+", **kwargs):
            pass

        return cls(fp, band=1)

    @property
    def dtype(self) -> np.dtype:
        return np.dtype(self.dataset.dtypes[self.band - 1])

    @property
    def height(self) -> int:
        """int : The number of rows in the raster."""  # noqa: D403
        return self.dataset.height  # type: ignore[no-any-return]

    @property
    def width(self) -> int:
        """int : The number of columns in the raster."""  # noqa: D403
        return self.dataset.width  # type: ignore[no-any-return]

    @property
    def shape(self):
        return self.height, self.width

    @property
    def closed(self) -> bool:
        """bool : True if the dataset is closed."""  # noqa: D403
        return self.dataset.closed  # type: ignore[no-any-return]

    def close(self) -> None:
        """Close the underlying dataset.

        Has no effect if the dataset is already closed.
        """
        if not self.closed:
            self.dataset.close()

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore[no-untyped-def]
        self.close()

    def _window_from_slices(self, key: slice | tuple[slice, ...]) -> Window:
        if isinstance(key, slice):
            row_slice = key
            col_slice = slice(None)
        else:
            row_slice, col_slice = key

        return Window.from_slices(
            row_slice, col_slice, height=self.height, width=self.width
        )

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}(dataset={self.dataset!r}, band={self.band!r})"

    def __setitem__(self, key: tuple[Index, ...], value: np.ndarray, /) -> None:
        with rasterio.open(
            self.filename,
            "r+",
        ) as dataset:
            if len(key) == 2:
                rows, cols = key
            elif len(key) == 3:
                _, rows, cols = _unpack_3d_slices(key)
            else:
                raise ValueError(
                    f"Invalid key for {self.__class__!r}.__setitem__: {key!r}"
                )
            try:
                window = Window.from_slices(
                    rows,
                    cols,
                    height=dataset.height,
                    width=dataset.width,
                )
            except rasterio.errors.WindowError as e:
                raise ValueError(f"Error creating window: {key = }, {value = }") from e
            return dataset.write(value, self.band, window=window)


class BackgroundRasterWriter(BackgroundWriter, DatasetWriter):
    """Class to write data to files in a background thread."""

    def __init__(
        self, filename: Filename, *, max_queue: int = 0, debug: bool = False, **kwargs
    ):
        super().__init__(nq=max_queue, name="Writer")
        if debug:
            #  background thread. Just synchronously write data
            self.notify_finished()
            self.queue_write = self.write  # type: ignore[assignment]

        if Path(filename).exists():
            self._raster = RasterWriter(filename)
        else:
            self._raster = RasterWriter.create(filename, **kwargs)
        self.filename = filename
        self.ndim = 2

    def write(self, key: tuple[Index, ...], value: np.ndarray):
        """Write out an ndarray to a subset of the pre-made `filename`.

        Parameters
        ----------
        key : tuple[Index,...]
            The key of the data to write.

        value : np.ndarray
            The block of data to write.

        """
        self._raster[key] = value

    def __setitem__(self, key: tuple[Index, ...], value: np.ndarray, /) -> None:
        self.queue_write(key, value)

    def close(self):
        """Close the underlying dataset and stop the background thread."""
        self._raster.close()
        self.notify_finished()

    @property
    def closed(self) -> bool:
        """bool : True if the dataset is closed."""  # noqa: D403
        return self._raster.closed

    @property
    def shape(self):
        return self._raster.shape

    @property
    def dtype(self) -> np.dtype:
        return self._raster.dtype

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore[no-untyped-def]
        self.close()


class BackgroundStackWriter(BackgroundWriter, DatasetStackWriter):
    """Class to write 3D data to multiple files in a background thread.

    Will create/overwrite the files in `file_list` if they exist.
    """

    def __init__(
        self,
        file_list: Sequence[Filename],
        *,
        like_filename: Filename | None = None,
        max_queue: int = 0,
        debug: bool = False,
        **file_creation_kwargs,
    ):
        from dolphin.io import write_arr

        super().__init__(nq=max_queue, name="GdalStackWriter")
        if debug:
            # Stop background thread. Just synchronously write data
            self.notify_finished()
            self.queue_write = self.write  # type: ignore[assignment]

        for fn in file_list:
            write_arr(
                arr=None,
                output_name=fn,
                like_filename=like_filename,
                **file_creation_kwargs,
            )

        self.file_list = file_list

        with rasterio.open(self.file_list[0]) as src:
            self.shape = (len(self.file_list), *src.shape)
            self.dtype = src.dtypes[0]

    def write(
        self, data: ArrayLike, row_start: int, col_start: int, band: int | None = None
    ):
        """Write out an ndarray to a subset of the pre-made `filename`.

        Parameters
        ----------
        data : ArrayLike
            3D data array to save.
        filename : Filename
            list of output files to save to, or (if cur_block is 2D) a single file.
        row_start : int
            Row index to start writing at.
        col_start : int
            Column index to start writing at.
        band : int, optional
            Band index in the file to write. Defaults to None, which uses first band,
            or whole raster for 3D data.

        Raises
        ------
        ValueError
            If length of `output_files` does not match length of `cur_block`.

        """
        from dolphin.io import write_block

        if data.ndim == 2:
            data = data[None, ...]
        if data.shape[0] != len(self.file_list):
            raise ValueError(f"{data.shape = }, but {len(self.file_list) = }")
        for fn, layer in zip(self.file_list, data):
            write_block(layer, fn, row_start, col_start, band=band)

    def __setitem__(self, key, value):
        # Unpack the slices
        band, rows, cols = _unpack_3d_slices(key)
        # Check we asked to write to all the files
        if band not in (slice(None), slice(None, None, None), ...):
            self.notify_finished()
            raise NotImplementedError("Can only write to all files at once.")
        self.queue_write(value, rows.start, cols.start)

    @property
    def closed(self) -> bool:
        """bool : True if the dataset is closed."""  # noqa: D403
        return self._thread.is_alive() is False

    def close(self) -> None:
        """Close the underlying dataset and stop the background thread."""
        self.notify_finished()



"""Module for simulating stacks of SLCs to test phase linking algorithms.

Contains simple versions of MLE and EVD estimator to compare against the
full CPU/GPU stack implementations.
"""

import numpy as np
import numpy.linalg as la
import scipy.ndimage as ndi
from numba import njit
from numpy.typing import ArrayLike


@njit(cache=True)
def _ccg_noise(N: int) -> np.array:
    return (np.random.randn(N) + 1j * np.random.randn(N)) / np.sqrt(2)


rng = np.random.default_rng()


@njit(cache=True)
def _seed(a):
    """Seed the random number generator for numba.

    https://numba.readthedocs.io/en/stable/reference/numpysupported.html#initialization
    """
    np.random.seed(a)


def ccg_noise(N: int) -> np.array:
    """Create N samples of standard complex circular Gaussian noise."""
    return (
        rng.normal(scale=1 / np.sqrt(2), size=2 * N)
        .astype(np.float32)
        .view(np.complex64)
    )


def simulate_neighborhood_stack(
    corr_matrix: ArrayLike, neighbor_samples: int = 200
) -> np.array:
    """Simulate one set of samples from a given correlation matrix.

    Parameters
    ----------
    corr_matrix : ArrayLike
        The correlation matrix to use for generating the samples.
    neighbor_samples : int
        The number of samples to generate for each pixel in the neighborhood.

    Returns
    -------
    np.array
        A 2D array of shape (N, neighbor_samples) containing the simulated samples,
        where N is the number of pixels in the neighborhood.

    """
    num_time = corr_matrix.shape[0]
    noise = ccg_noise(neighbor_samples * num_time)
    noise_as_columns = noise.reshape(num_time, neighbor_samples)

    L = np.linalg.cholesky(corr_matrix)
    samps = L @ noise_as_columns
    return samps


def simulate_coh(
    num_acq: int = 50,
    gamma_inf: float = 0.1,
    gamma0: float = 0.999,
    Tau0: float = 72,
    acq_interval: int = 12,
    add_signal: bool = False,
    signal_std: float = 0.1,
) -> np.array:
    """Simulate a correlation matrix for a pixel.

    Parameters
    ----------
    num_acq : int
        The number of acquisitions.
    gamma_inf : float
        The asymptotic coherence value.
    gamma0 : float
        The initial coherence value.
    Tau0 : float
        The coherence decay time constant.
    acq_interval : int
        The acquisition interval in days.
    add_signal : bool
        Whether to add a simulated signal.
    signal_std : float
        The standard deviation of the simulated signal.

    Returns
    -------
    np.array
        The simulated correlation matrix.
    np.array
        The simulated truth signal.

    """
    time_series_length = num_acq * acq_interval
    time = np.arange(0, time_series_length, acq_interval)
    if add_signal:
        k, signal_rate = 1, 2
        noisy_truth, truth = _sim_signal(
            time,
            signal_rate=signal_rate,
            std_random=signal_std,
            k=k,
        )
    else:
        noisy_truth = truth = np.zeros(len(time), dtype=np.float64)

    C = simulate_coh_stack(time, gamma0, gamma_inf, Tau0, noisy_truth).squeeze()
    return C, truth


def simulate_coh_stack(
    time: np.ndarray,
    gamma0: np.ndarray,
    gamma_inf: np.ndarray,
    Tau0: np.ndarray,
    signal: np.ndarray,
    amplitudes: ArrayLike | None = None,
) -> np.ndarray:
    """Create a coherence matrix at each pixel.

    Parameters
    ----------
    time : np.ndarray
        Array of time values, arbitrary units.
        shape (N,) where N is the number of acquisitions.
    gamma0 : np.ndarray
        Initial coherence value for each pixel.
    gamma_inf : np.ndarray
        Asymptotic coherence value for each pixel.
    Tau0 : np.ndarray
        Coherence decay time constant for each pixel.
    signal : np.ndarray
        Simulated signal phase for each pixel.
    amplitudes: ArrayLike, optional
        If provided, set the amplitudes of the output pixels.
        Default is to use all ones.

    Returns
    -------
    np.ndarray
        The simulated coherence matrix for each pixel.

    """
    num_time = time.shape[0]
    time_diff = time[:, None] - time[None, :]
    time_diff = time_diff[None, None, :, :]
    gamma0 = np.atleast_2d(gamma0)[:, :, None, None]
    gamma_inf = np.atleast_2d(gamma_inf)[:, :, None, None]
    Tau0 = np.atleast_2d(Tau0)[:, :, None, None]
    if amplitudes is None:
        amplitudes = np.ones((num_time, 1, 1), dtype=np.float32)

    gamma = (gamma0 - gamma_inf) * np.exp(time_diff / Tau0) + gamma_inf
    phase_diff = signal[:, None] - signal[None, :]
    # Multiply each pixel by (a a^T)
    A = np.einsum("irc,jrc->rcij", amplitudes, amplitudes)
    C = A * gamma * np.exp(1j * phase_diff)

    rl, cl = np.tril_indices(num_time, k=-1)

    C[..., rl, cl] = np.conj(np.transpose(C, axes=(0, 1, 3, 2))[..., rl, cl])

    # Reset the diagonals of each pixel to 1
    rs, cs = np.diag_indices(num_time)
    C[:, :, rs, cs] = 1
    return C


def _sim_signal(
    t: np.ndarray,
    signal_rate: float = 1.0,
    std_random: float = 0,
    k: int = 1,
) -> tuple[np.ndarray, np.ndarray]:
    """Simulate a signal with a linear trend, seasonal component, and random noise.

    Parameters
    ----------
    t : np.ndarray
        Time array in days.
    signal_rate : float, optional
        Linear rate of the signal in radians per year, by default 1.0.
    std_random : float, optional
        Standard deviation of the random noise component, by default 0.
    k : int, optional
        Seasonal parameter, 1 for annual and 2 for semi-annual, by default 1.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        A tuple containing the simulated signal phase and the true signal.

    """
    truth = signal_rate * (t - t[0]) / 365.0
    if k > 0:
        seasonal = np.sin(2 * np.pi * k * t / 365.0) + np.cos(2 * np.pi * k * t / 365.0)
        truth += seasonal

    # adding random temporal signal (which simulates atmosphere + DEM error + ...)
    signal_phase = truth + std_random / 2 * np.random.randn(len(t))
    # we divided std by 2 since we're subtracting the first value
    signal_phase = signal_phase - signal_phase[0]

    # wrap the signal_phase to -pi to pi
    signal_phase = np.mod(signal_phase + np.pi, 2 * np.pi) - np.pi
    truth = np.mod((truth - truth[0]) + np.pi, 2 * np.pi) - np.pi

    return signal_phase.astype(np.float64), truth.astype(np.float64)


def rmse(x, y, axis=None):
    """Calculate the root mean squared error between two arrays.

    If x and y are complex, the RMSE is calculated using the angle between them.
    """
    if np.iscomplexobj(x) and np.iscomplexobj(y):
        return np.sqrt(np.mean((np.angle(x * y.conj()) ** 2), axis=axis))

    return np.sqrt(np.mean((x - y) ** 2, axis=axis))


@njit(cache=True)
def mle(cov_mat, beta=0.00):
    """Estimate the linked phase using the MLE estimator.

    Parameters
    ----------
    cov_mat : np.array
        The sample covariance matrix
    beta : float, optional
        The regularization parameter, by default 0.0

    Returns
    -------
    np.array
        The estimated linked phase

    """
    dtype = cov_mat.dtype
    cov_mat = cov_mat.astype(dtype)
    # estimate the wrapped phase based on the EMI paper
    # *smallest* eigenvalue decomposition of the (|Gamma|^-1  *  C) matrix
    Gamma = np.abs(cov_mat).astype(dtype)
    if beta:
        Gamma = (1 - beta) * Gamma + beta * np.eye(Gamma.shape[0], dtype=Gamma.dtype)
        Gamma_inv = la.inv(Gamma).astype(dtype)
    else:
        Gamma_inv = la.inv(Gamma).astype(dtype)
    _, v = la.eigh(Gamma_inv * cov_mat)

    # smallest eigenvalue is idx 0
    # reference to the first acquisition
    evd_estimate = v[:, 0] * np.conjugate(v[0, 0])
    return evd_estimate.astype(dtype)


@njit(cache=True)
def evd(cov_mat):
    """Estimate the linked phase the largest eigenvector of `cov_mat`."""
    # estimate the wrapped phase based on the eigenvalue decomp of the cov. matrix
    # n = len(cov_mat)
    # lambda_, v = la.eigh(cov_mat, subset_by_index=[n - 1, n - 1])  # only scipy.linalg
    # v = v.flatten()
    lambda_, v = la.eigh(cov_mat)

    # Biggest eigenvalue is the last one
    # reference to the first acquisition
    evd_estimate = v[:, -1] * np.conjugate(v[0, -1])

    return evd_estimate.astype(cov_mat.dtype)


def make_defo_stack(
    shape: tuple[int, int, int], sigma: float, max_amplitude: float = 1
) -> np.ndarray:
    """Create the time series of deformation to add to each SAR date.

    Parameters
    ----------
    shape : tuple[int, int, int]
        Shape of the deformation stack (num_time_steps, rows, cols).
    sigma : float
        Standard deviation of the Gaussian deformation.
    max_amplitude : float, optional
        Maximum amplitude of the final deformation. Defaults to 1.

    Returns
    -------
    np.ndarray
        Deformation stack with time series, shape (num_time_steps, rows, cols).

    """
    num_time_steps, rows, cols = shape
    shape2d = (rows, cols)
    # Get shape of deformation in final form (normalized to 1 max)
    final_defo = make_gaussian(shape=shape2d, sigma=sigma).reshape((1, *shape2d))
    final_defo *= max_amplitude / np.max(final_defo)
    # Broadcast this shape with linear evolution
    time_evolution = np.linspace(0, 1, num=num_time_steps)[:, None, None]
    return (final_defo * time_evolution).astype(np.float32)


def make_noisy_samples(
    C: ArrayLike,
    defo_stack: ArrayLike,
) -> np.ndarray:
    """Create noisy deformation samples given a covariance matrix and deformation stack.

    Parameters
    ----------
    C : ArrayLike,
        Covariance matrix of shape (num_time, num_time) , or you can pass
        on matrix per pixel as shape (rows, cols, num_time, num_time).
    defo_stack : ArrayLike,
        Deformation stack of shape (num_time, rows, cols).

    Returns
    -------
    samps3d: np.ndarray
        Noisy deformation samples of shape (num_time, rows, cols).

    """

    def _get_diffs(stack: np.ndarray) -> np.ndarray:
        """Create all differences between the deformation stack.

        Parameters
        ----------
        stack : np.ndarray
            Signal stack of shape (num_time, rows, cols).

        Returns
        -------
        np.ndarray, complex64
            Covariance phases of shape (rows, cols, num_time, num_time).

        """
        # Create all possible differences using broadcasting
        # shape: (num_time, 1, rows, cols)
        stack_i = np.exp(1j * stack[:, None, :, :])
        # shape: (1, num_time, rows, cols)
        stack_j = np.exp(1j * stack[None, :, :, :])
        diff_stack = stack_i * stack_j.conj()

        # Reshape to (rows, cols, num_time, num_time)
        return diff_stack.transpose(2, 3, 0, 1)

    num_time, rows, cols = defo_stack.shape
    shape2d = (rows, cols)
    num_pixels = np.prod(shape2d)

    C_tiled = np.tile(C, (*shape2d, 1, 1)) if C.squeeze().ndim == 2 else C

    if C_tiled.shape != (rows, cols, num_time, num_time):
        raise ValueError(f"{C_tiled.shape=}, but {defo_stack.shape=}")
    # Reset the diagonals of each pixel to 1
    rs, cs = np.diag_indices(num_time)
    C_tiled[:, :, rs, cs] = 1

    signal_cov = _get_diffs(defo_stack)
    C_tiled_with_signal = C_tiled * signal_cov
    C_unstacked = C_tiled_with_signal.reshape(num_pixels, num_time, num_time)

    noise = ccg_noise(num_time * num_pixels)
    noise_unstacked = noise.reshape(num_pixels, num_time, 1)

    L_unstacked = np.linalg.cholesky(C_unstacked)
    samps = L_unstacked @ noise_unstacked

    samps3d = samps.reshape(*shape2d, num_time)
    return np.moveaxis(samps3d, -1, 0)


def make_gaussian(
    shape: tuple[int, int],
    sigma: float,
    row: int | None = None,
    col: int | None = None,
    normalize: bool = False,
    amp: float | None = None,
    noise_sigma: float = 0.0,
) -> np.ndarray:
    """Create a Gaussian blob of given shape and width.

    Parameters
    ----------
    shape : tuple[int, int]
        (rows, cols)
    sigma : float
        Standard deviation of the Gaussian.
    row : int, optional
        Center row of the blob. Defaults to None.
    col : int, optional
        Center column of the blob. Defaults to None.
    normalize : bool, optional
        Normalize the amplitude peak to 1. Defaults to False.
    amp : float, optional
        Peak height of the Gaussian. Defaults to None.
    noise_sigma : float, optional
        Standard deviation of random Gaussian noise added to the image. Defaults to 0.0.

    Returns
    -------
    ndarray
        Gaussian blob.

    """
    delta = np.zeros(shape)
    rows, cols = shape
    if col is None:
        col = cols // 2
    if row is None:
        row = rows // 2
    delta[row, col] = 1

    out = ndi.gaussian_filter(delta, sigma, mode="constant") * sigma**2
    if normalize or amp is not None:
        out /= out.max()
    if amp is not None:
        out *= amp
    if noise_sigma > 0:
        out += noise_sigma * np.random.standard_normal(shape)
    return out



import os
from math import floor

import numpy as np
import numpy.testing as npt
import pytest

from dolphin._types import HalfWindow, Strides
from dolphin.phase_link import covariance, simulate
from dolphin.utils import gpu_is_available, take_looks

GPU_AVAILABLE = gpu_is_available() and os.environ.get("NUMBA_DISABLE_JIT") != "1"
NUM_ACQ = 30
simulate._seed(1234)


# Make sure the GPU versions are correct by making simpler versions:
def form_cov(slc1, slc2, looks):
    numerator = take_looks(slc1 * slc2.conj(), *looks)
    amplitude1 = take_looks(slc1 * slc1.conj(), *looks)
    amplitude2 = take_looks(slc2 * slc2.conj(), *looks)
    return numerator / np.sqrt(amplitude1 * amplitude2)


def get_expected_cov(slcs, looks):
    # Manually (slowly) form the covariance matrix at all pixels
    N = slcs.shape[0]
    rows, cols = take_looks(slcs[0], *looks).shape
    out = np.zeros((rows, cols, N, N), dtype=np.complex64)
    for i in range(N):
        for j in range(i + 1, N):
            out[:, :, i, j] = form_cov(slcs[i], slcs[j], looks)
            out[:, :, j, i] = out[:, :, i, j].conj()
        out[:, :, i, i] = 1.0
    return out


@pytest.fixture(scope="module")
def slcs(shape=(10, 100, 100)):
    return np.random.rand(*shape) + 1j * np.random.rand(*shape)


@pytest.fixture()
def expected_cov(slcs, looks=(5, 5)):
    return get_expected_cov(slcs, looks)


def test_coh_mat_single(slcs, expected_cov, looks=(5, 5)):
    num_slc, rows, cols = slcs.shape

    # Check the single pixel function
    expected_looked_size = tuple(
        floor(size / look) for size, look in zip((rows, cols), looks)
    )
    assert expected_cov.shape == ((*expected_looked_size, num_slc, num_slc))

    r_looks, c_looks = looks
    for r in range(expected_looked_size[0]):
        for c in range(expected_looked_size[1]):
            r_slice = slice(r * r_looks, (r + 1) * r_looks)
            c_slice = slice(c * c_looks, (c + 1) * c_looks)
            cur_samples = slcs[:, r_slice, c_slice].reshape(num_slc, -1)
            cur_C = covariance.coh_mat_single(cur_samples)
            npt.assert_array_almost_equal(expected_cov[r, c, :, :], cur_C)


def test_estimate_stack_covariance(slcs, expected_cov, looks=(5, 5)):
    # Check the full stack function
    r_looks, c_looks = looks
    half_window = HalfWindow(c_looks // 2, r_looks // 2)
    strides = Strides(c_looks, r_looks)
    C1 = covariance.estimate_stack_covariance(
        slcs, half_window=half_window, strides=strides
    )
    npt.assert_array_almost_equal(expected_cov, C1)


def test_estimate_stack_covariance_nans_pixel(slcs):
    num_slc, _, _ = slcs.shape

    C = covariance.coh_mat_single(slcs.reshape(num_slc, -1))
    # Nans for one pixel in all SLCs
    slc_stack_nan = slcs.copy()
    slc_stack_nan[:, 1, 1] = np.nan
    slc_samples_nan = slc_stack_nan.reshape(num_slc, -1)
    C_nan = covariance.coh_mat_single(slc_samples_nan)
    assert np.max(np.abs(C - C_nan)) < 0.01


def test_estimate_stack_covariance_nans_image(slcs):
    num_slc, _, _ = slcs.shape
    # Nans for an entire SLC
    slc_stack_nan = slcs.copy()
    slc_stack_nan[1, :, :] = np.nan
    slc_samples_nan = slc_stack_nan.reshape(num_slc, -1)
    # TODO: should we raise an error if we pass a dead SLC?
    assert covariance.coh_mat_single(slc_samples_nan)[0, 1] == 0
    assert covariance.coh_mat_single(slc_samples_nan)[1, 0] == 0


def test_estimate_stack_covariance_all_nans(slcs):
    num_slc, _, _ = slcs.shape
    # all nans should return an identity matrix (no correlation anywhere)
    slc_stack_nan = slcs.copy()
    slc_stack_nan[:, :, :] = np.nan
    slc_samples_nan = slc_stack_nan.reshape(num_slc, -1)
    # TODO: do we want an identity matrix? Or for it to be 0?
    assert covariance.coh_mat_single(slc_samples_nan).sum() == 0
    # assert np.isnan(covariance.coh_mat_single(slc_samples_nan)).all()


def test_estimate_stack_covariance_neighbors(slcs):
    num_slc, rows, cols = slcs.shape

    C = covariance.coh_mat_single(slcs.reshape(num_slc, -1))
    # We're checking that setting the pixels to nan
    # is the same as saying they are not neighbors

    # Nans for one pixel in all SLCs
    slc_stack_nan = slcs.copy()
    nan_row, nan_col = (1, 1)
    slc_stack_nan[:, nan_row, nan_col] = np.nan

    slc_samples_nan = slc_stack_nan.reshape(num_slc, -1)
    C_nan = covariance.coh_mat_single(slc_samples_nan)

    slc_samples = slcs.reshape(num_slc, -1)
    neighbor_mask = np.ones(slc_samples.shape[1], dtype=np.bool_)
    idx = np.ravel_multi_index((nan_row, nan_col), (rows, cols))
    neighbor_mask[idx] = False

    C_neighbors = covariance.coh_mat_single(slc_samples, neighbor_mask=neighbor_mask)
    npt.assert_allclose(C_nan, C_neighbors)

    with pytest.raises(AssertionError):
        # Make sure this is different than the original
        npt.assert_allclose(C, C_neighbors)


def test_estimate_stack_covariance_neighbors_masked(slcs):
    num_slc, rows, cols = slcs.shape
    # Now mask an entire row
    slc_stack_nan = slcs.copy()
    slc_stack_nan[:, 0, :] = np.nan
    slc_samples_nan = slc_stack_nan.reshape(num_slc, -1)
    C_nan = covariance.coh_mat_single(slc_samples_nan)

    slc_samples = slcs.reshape(num_slc, -1)
    neighbor_mask = np.ones(slc_samples.shape[1], dtype=np.bool_)
    neighbor_mask[:] = True
    neighbor_mask[:cols] = False

    C_neighbors = covariance.coh_mat_single(slc_samples, neighbor_mask=neighbor_mask)
    npt.assert_allclose(C_nan, C_neighbors)



from __future__ import annotations

import numba
import numpy as np
from numpy.typing import ArrayLike


@numba.njit(nogil=True)
def remove_unconnected(data: ArrayLike, inplace: bool = True) -> np.ndarray:
    """Remove groups of True values in a boolean matrix not connected to the center.

    Parameters
    ----------
    data : ArrayLike
        2D input boolean matrix.
    inplace : bool, default True
        If True, modifies the input array directly. Otherwise, creates a copy.

    Returns
    -------
    np.ndarray
        Boolean matrix with only the connected group of True values centered
        around the middle pixel.

    Notes
    -----
    This function considers the 8 surrounding neighbors as connected
        (i.e. includes diagonals.)

    """
    if not inplace:
        data = data.copy()
    rows, cols = data.shape
    visited = np.zeros((rows, cols), dtype=np.bool_)
    # Using a stack instead of recursive calls for speed
    # Pre-allocate the stack
    max_stack_size = 1 + rows * cols
    stack = np.zeros((3 * max_stack_size, 2), dtype=np.int16)
    stack_ptr = 0  # Stack pointer to keep track of the top of the stack

    # Start at the center and search outward.
    start_row, start_col = rows // 2, cols // 2
    visited[start_row, start_col] = True
    stack[stack_ptr] = start_row, start_col
    stack_ptr += 1

    connected_idxs = [
        (-1, -1),
        (-1, 0),
        (-1, 1),
        (0, -1),
        (0, 1),
        (1, -1),
        (1, 0),
        (1, 1),
    ]

    # Run the Depth First Search to get connected pixels
    while stack_ptr > 0:
        stack_ptr -= 1
        x, y = stack[stack_ptr]

        for dx, dy in connected_idxs:
            x2, y2 = x + dx, y + dy
            # Skip OOB pixels
            if x2 < 0 or x2 >= rows or y2 < 0 or y2 >= cols:
                continue
            # Check we haven't visited:
            if visited[x2, y2]:
                continue
            # Now push to stack if this pixel is True
            if data[x2, y2]:
                visited[x2, y2] = True
                stack[stack_ptr] = x2, y2
                stack_ptr += 1

    # Finally only keep ones from the original that were visited
    data &= visited
    return data



"""Stitch burst interferograms (optional) and unwrap them."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Sequence

from dolphin import stitching
from dolphin._log import log_runtime
from dolphin._overviews import ImageType, create_image_overviews, create_overviews
from dolphin._types import Bbox
from dolphin.interferogram import estimate_interferometric_correlations
from dolphin.io._utils import repack_raster

from .config import OutputOptions

logger = logging.getLogger(__name__)


@log_runtime
def run(
    ifg_file_list: Sequence[Path],
    temp_coh_file_list: Sequence[Path],
    ps_file_list: Sequence[Path],
    amp_dispersion_list: Sequence[Path],
    shp_count_file_list: Sequence[Path],
    stitched_ifg_dir: Path,
    output_options: OutputOptions,
    file_date_fmt: str = "%Y%m%d",
    corr_window_size: tuple[int, int] = (11, 11),
    num_workers: int = 3,
) -> tuple[list[Path], list[Path], Path, Path, Path, Path]:
    """Run the displacement workflow on a stack of SLCs.

    Parameters
    ----------
    ifg_file_list : Sequence[Path]
        Sequence of burst-wise interferograms files to stitch.
    temp_coh_file_list : Sequence[Path]
        Sequence of paths to the burst-wise temporal coherence files.
    ps_file_list : Sequence[Path]
        Sequence of paths to the (looked) burst-wise ps mask files.
    amp_dispersion_list : Sequence[Path]
        Sequence of paths to the (looked) burst-wise amplitude dispersion files.
    shp_count_file_list : Sequence[Path]
        Sequence of paths to the burst-wise SHP counts files.
    stitched_ifg_dir : Path
        Location to store the output stitched ifgs and correlations
    output_options : OutputOptions
        [`UnwrapWorkflow`][dolphin.workflows.config.OutputOptions] object
        for with parameters for the input/output options
    file_date_fmt : str
        Format of dates contained in filenames.
        default = "%Y%m%d"
    corr_window_size : tuple[int, int]
        Size of moving window (rows, cols) to use for estimating correlation.
        Default = (11, 11)
    num_workers : int
        Number of threads to use for stitching in parallel.
        Default = 3

    Returns
    -------
    stitched_ifg_paths : list[Path]
        list of Paths to the stitched interferograms.
    interferometric_corr_paths : list[Path]
        list of Paths to interferometric correlation files created.
    stitched_temp_coh_file : Path
        Path to temporal correlation file created.
    stitched_ps_file : Path
        Path to ps mask file created.
    stitched_amp_disp_file : Path
        Path to amplitude dispersion file created.
    stitched_shp_count_file : Path
        Path to SHP count file created.

    """
    stitched_ifg_dir.mkdir(exist_ok=True, parents=True)
    # Also preps for snaphu, which needs binary format with no nans
    logger.info("Stitching interferograms by date.")
    out_bounds = Bbox(*output_options.bounds) if output_options.bounds else None
    date_to_ifg_path = stitching.merge_by_date(
        image_file_list=ifg_file_list,
        file_date_fmt=file_date_fmt,
        output_dir=stitched_ifg_dir,
        output_suffix=".int.tif",
        driver="GTiff",
        out_bounds=out_bounds,
        out_bounds_epsg=output_options.bounds_epsg,
        num_workers=num_workers,
    )
    stitched_ifg_paths = list(date_to_ifg_path.values())

    # Estimate the interferometric correlation from the stitched interferogram
    interferometric_corr_paths = estimate_interferometric_correlations(
        stitched_ifg_paths,
        window_size=corr_window_size,
        num_workers=num_workers,
    )

    # Stitch the correlation files
    stitched_temp_coh_file = stitched_ifg_dir / "temporal_coherence.tif"
    if not stitched_temp_coh_file.exists():
        stitching.merge_images(
            temp_coh_file_list,
            outfile=stitched_temp_coh_file,
            driver="GTiff",
            out_bounds=out_bounds,
            out_bounds_epsg=output_options.bounds_epsg,
        )
        repack_raster(stitched_temp_coh_file, keep_bits=10)

    # Stitch the looked PS files
    stitched_ps_file = stitched_ifg_dir / "ps_mask_looked.tif"
    if not stitched_ps_file.exists():
        stitching.merge_images(
            ps_file_list,
            outfile=stitched_ps_file,
            out_nodata=255,
            driver="GTiff",
            resample_alg="nearest",
            out_bounds=out_bounds,
            out_bounds_epsg=output_options.bounds_epsg,
        )

    # Stitch the amp dispersion files
    stitched_amp_disp_file = stitched_ifg_dir / "amp_dispersion_looked.tif"
    if not stitched_amp_disp_file.exists():
        stitching.merge_images(
            amp_dispersion_list,
            outfile=stitched_amp_disp_file,
            driver="GTiff",
            out_bounds=out_bounds,
            out_bounds_epsg=output_options.bounds_epsg,
        )
        repack_raster(stitched_temp_coh_file, keep_bits=10)

    stitched_shp_count_file = stitched_ifg_dir / "shp_counts.tif"
    stitching.merge_images(
        shp_count_file_list,
        outfile=stitched_shp_count_file,
        driver="GTiff",
        out_bounds=out_bounds,
        out_bounds_epsg=output_options.bounds_epsg,
    )

    if output_options.add_overviews:
        logger.info("Creating overviews for stitched images")
        create_overviews(stitched_ifg_paths, image_type=ImageType.INTERFEROGRAM)
        create_overviews(interferometric_corr_paths, image_type=ImageType.CORRELATION)
        create_image_overviews(stitched_ps_file, image_type=ImageType.PS)
        create_image_overviews(stitched_temp_coh_file, image_type=ImageType.CORRELATION)
        create_image_overviews(stitched_amp_disp_file, image_type=ImageType.CORRELATION)
        create_image_overviews(stitched_shp_count_file, image_type=ImageType.PS)

    return (
        stitched_ifg_paths,
        interferometric_corr_paths,
        stitched_temp_coh_file,
        stitched_ps_file,
        stitched_amp_disp_file,
        stitched_shp_count_file,
    )
